# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.backend import ops
from keras_cv.src.backend import random


def balanced_sample(
    positive_matches,
    negative_matches,
    num_samples: int,
    positive_fraction: float,
):
    """
    Sampling ops to balance positive and negative samples, deals with both
    batched and unbatched inputs.

    Args:
      positive_matches: [N] or [batch_size, N] boolean Tensor, True for
        indicating the index is a positive sample
      negative_matches: [N] or [batch_size, N] boolean Tensor, True for
        indicating the index is a negative sample
      num_samples: int, representing the number of samples to collect
      positive_fraction: float. 0.5 means positive samples should be half
        of all collected samples.

    Returns:
      selected_indicators: [N] or [batch_size, N]
        integer Tensor, 1 for indicating the index is sampled, 0 for
        indicating the index is not sampled.
    """

    N = ops.shape(positive_matches)[-1]
    if N < num_samples:
        raise ValueError(
            "passed in {positive_matches.shape} has less element than "
            f"{num_samples}"
        )
    # random_val = tf.random.uniform(tf.shape(positive_matches), minval=0.,
    # maxval=1.)
    zeros = ops.zeros_like(positive_matches, dtype="float32")
    ones = ops.ones_like(positive_matches, dtype="float32")
    ones_rand = ones + random.uniform(ops.shape(ones), minval=-0.2, maxval=0.2)
    halfs = 0.5 * ops.ones_like(positive_matches, dtype="float32")
    halfs_rand = halfs + random.uniform(
        ops.shape(halfs), minval=-0.2, maxval=0.2
    )
    values = zeros
    values = ops.where(positive_matches, ones_rand, values)
    values = ops.where(negative_matches, halfs_rand, values)
    num_pos_samples = int(num_samples * positive_fraction)
    valid_matches = ops.logical_or(positive_matches, negative_matches)
    # this might contain negative samples as well
    _, positive_indices = ops.top_k(values, k=num_pos_samples)
    selected_indicators = ops.cast(
        ops.sum(ops.one_hot(positive_indices, N), axis=-2), dtype="bool"
    )
    # setting all selected samples to zeros
    values = ops.where(selected_indicators, zeros, values)
    # setting all excessive positive matches to zeros as well
    values = ops.where(positive_matches, zeros, values)
    num_neg_samples = num_samples - num_pos_samples
    _, negative_indices = ops.top_k(values, k=num_neg_samples)
    selected_indices = ops.concatenate(
        [positive_indices, negative_indices], axis=-1
    )
    selected_indicators = ops.sum(ops.one_hot(selected_indices, N), axis=-2)
    selected_indicators = ops.minimum(
        selected_indicators, ops.ones_like(selected_indicators)
    )
    selected_indicators = ops.where(
        valid_matches, selected_indicators, ops.zeros_like(selected_indicators)
    )
    return selected_indicators



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers import preprocessing
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils


@keras_cv_export("keras_cv.layers.RandomColorJitter")
class RandomColorJitter(VectorizedBaseImageAugmentationLayer):
    """RandomColorJitter class randomly apply brightness, contrast, saturation
    and hue image processing operation sequentially and randomly on the
    input. It expects input as RGB image. The expected image should be
    `(0-255)` pixel ranges.

    Input shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `channels_last` format
    Output shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `channels_last` format

    Args:
        value_range: the range of values the incoming images will have.
            Represented as a two number tuple written [low, high].
            This is typically either `[0, 1]` or `[0, 255]` depending
            on how your preprocessing pipeline is set up.
        brightness_factor: Float or a list/tuple of 2 floats between -1.0
            and 1.0. The factor is used to determine the lower bound and
            upper bound of the brightness adjustment. A float value will be
            chosen randomly between the limits. When -1.0 is chosen, the
            output image will be black, and when 1.0 is chosen, the image
            will be fully white. When only one float is provided, eg, 0.2,
            then -0.2 will be used for lower bound and 0.2 will be used for
            upper bound.
        contrast_factor: A positive float represented as fraction of value,
            or a tuple of size 2 representing lower and upper bound. When
            represented as a single float, lower = upper. The contrast factor
            will be randomly picked between `[1.0 - lower, 1.0 + upper]`.
        saturation_factor: Either a tuple of two floats or a single float.
            `factor` controls the extent to which the image saturation is
            impacted. `factor=0.5` makes this layer perform a no-op operation.
            `factor=0.0` makes the image to be fully grayscale. `factor=1.0`
            makes the image to be fully saturated.
        hue_factor: A tuple of two floats, a single float or
            `keras_cv.FactorSampler`. `factor` controls the extent to which the
            image sharpness is impacted. `factor=0.0` makes this layer perform
            a no-op operation, while a value of 1.0 performs the most aggressive
            contrast adjustment available. If a tuple is used, a `factor` is
            sampled between the two values for every image augmented. If a
            single float is used, a value between `0.0` and the passed float is
            sampled. In order to ensure the value is always the same, please
            pass a tuple with two identical floats: `(0.5, 0.5)`.
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    color_jitter = keras_cv.layers.RandomColorJitter(
            value_range=(0, 255),
            brightness_factor=(-0.2, 0.5),
            contrast_factor=(0.5, 0.9),
            saturation_factor=(0.5, 0.9),
            hue_factor=(0.5, 0.9),
    )
    augmented_images = color_jitter(images)
    ```
    """

    def __init__(
        self,
        value_range,
        brightness_factor,
        contrast_factor,
        saturation_factor,
        hue_factor,
        seed=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.value_range = value_range
        self.brightness_factor = brightness_factor
        self.contrast_factor = contrast_factor
        self.saturation_factor = saturation_factor
        self.hue_factor = hue_factor
        self.seed = seed

        self.random_brightness = preprocessing.RandomBrightness(
            factor=self.brightness_factor, value_range=(0, 255), seed=self.seed
        )
        self.random_contrast = preprocessing.RandomContrast(
            factor=self.contrast_factor, value_range=(0, 255), seed=self.seed
        )
        self.random_saturation = preprocessing.RandomSaturation(
            factor=self.saturation_factor, seed=self.seed
        )
        self.random_hue = preprocessing.RandomHue(
            factor=self.hue_factor, value_range=(0, 255), seed=self.seed
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            images=image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations=None, **kwargs):
        images = preprocessing_utils.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
            dtype=self.compute_dtype,
        )
        images = self.random_brightness(images)
        images = self.random_contrast(images)
        images = self.random_saturation(images)
        images = self.random_hue(images)
        images = preprocessing_utils.transform_value_range(
            images,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "brightness_factor": self.brightness_factor,
                "contrast_factor": self.contrast_factor,
                "saturation_factor": self.saturation_factor,
                "hue_factor": self.hue_factor,
                "seed": self.seed,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from tensorflow.keras.callbacks import Callback

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import ops
from keras_cv.src.metrics.coco import compute_pycoco_metrics
from keras_cv.src.models.object_detection.__internal__ import unpack_input
from keras_cv.src.utils.conditional_imports import assert_pycocotools_installed


@keras_cv_export("keras_cv.callbacks.PyCOCOCallback")
class PyCOCOCallback(Callback):
    def __init__(
        self, validation_data, bounding_box_format, cache=True, **kwargs
    ):
        """Creates a callback to evaluate PyCOCO metrics on a validation
        dataset.

        Args:
            validation_data: a tf.data.Dataset containing validation data.
                Entries should have the form ```(images, {"boxes": boxes,
                "classes": classes})```.
            bounding_box_format: the KerasCV bounding box format used in the
                validation dataset (e.g. "xywh")
            cache: whether the callback should cache the dataset between
                iterations. Note that if the validation dataset has shuffling of
                any kind (e.g. from `shuffle_files=True` in a call to TFDS).
                Load or a call to tf.data.Dataset.shuffle() with
                `reshuffle_each_iteration=True`), you **must** cache the dataset
                to preserve iteration order. This will store your entire dataset
                in main memory, so for large datasets consider avoiding shuffle
                operations and passing `cache=False`.
        """
        assert_pycocotools_installed("PyCOCOCallback")
        self.val_data = validation_data
        if cache:
            # We cache the dataset to preserve a consistent iteration order.
            self.val_data = self.val_data.cache()
        self.bounding_box_format = bounding_box_format
        super().__init__(**kwargs)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}

        def images_only(data, maybe_boxes=None):
            if maybe_boxes is None:
                images, boxes = unpack_input(data)
            else:
                images = data
            return images

        def boxes_only(data, maybe_boxes=None):
            if maybe_boxes is None:
                images, boxes = unpack_input(data)
            else:
                boxes = maybe_boxes
            return boxes

        images_only_ds = self.val_data.map(images_only)
        y_pred = self.model.predict(images_only_ds)
        box_pred = y_pred["boxes"]
        cls_pred = ops.convert_to_numpy(y_pred["classes"])
        confidence_pred = ops.convert_to_numpy(y_pred["confidence"])
        valid_det = ops.convert_to_numpy(y_pred["num_detections"])

        gt = [boxes for boxes in self.val_data.map(boxes_only)]
        gt_boxes = ops.concatenate(
            [ops.convert_to_numpy(boxes["boxes"]) for boxes in gt],
            axis=0,
        )
        gt_classes = ops.concatenate(
            [ops.convert_to_numpy(boxes["classes"]) for boxes in gt],
            axis=0,
        )

        first_image_batch = next(iter(images_only_ds))
        height = first_image_batch.shape[1]
        width = first_image_batch.shape[2]
        total_images = gt_boxes.shape[0]

        gt_boxes = bounding_box.convert_format(
            gt_boxes, source=self.bounding_box_format, target="yxyx"
        )

        source_ids = np.char.mod(
            "%d", np.linspace(1, total_images, total_images)
        )
        num_detections = ops.sum(ops.cast(gt_classes > 0, "int32"), axis=-1)

        ground_truth = {
            "source_id": [source_ids],
            "height": [
                ops.convert_to_numpy(
                    ops.tile(ops.array([height]), [total_images])
                )
            ],
            "width": [
                ops.convert_to_numpy(
                    ops.tile(ops.array([width]), [total_images])
                )
            ],
            "num_detections": [ops.convert_to_numpy(num_detections)],
            "boxes": [ops.convert_to_numpy(gt_boxes)],
            "classes": [ops.convert_to_numpy(gt_classes)],
        }

        box_pred = bounding_box.convert_format(
            box_pred, source=self.bounding_box_format, target="yxyx"
        )

        predictions = {
            "source_id": [source_ids],
            "detection_boxes": [ops.convert_to_numpy(box_pred)],
            "detection_classes": [cls_pred],
            "detection_scores": [confidence_pred],
            "num_detections": [valid_det],
        }

        metrics = compute_pycoco_metrics(ground_truth, predictions)
        # Mark these as validation metrics by prepending a val_ prefix
        metrics = {"val_" + name: val for name, val in metrics.items()}

        logs.update(metrics)



# Copyright 2022 Waymo LLC.
#
# Licensed under the terms in https://github.com/keras-team/keras-cv/blob/master/keras_cv/layers/preprocessing_3d/waymo/LICENSE  # noqa: E501

import tensorflow as tf

from keras_cv.src import point_cloud
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import random
from keras_cv.src.layers.preprocessing_3d import base_augmentation_layer_3d

POINT_CLOUDS = base_augmentation_layer_3d.POINT_CLOUDS
BOUNDING_BOXES = base_augmentation_layer_3d.BOUNDING_BOXES
POINTCLOUD_LABEL_INDEX = base_augmentation_layer_3d.POINTCLOUD_LABEL_INDEX


@keras_cv_export("keras_cv.layers.FrustumRandomDroppingPoints")
class FrustumRandomDroppingPoints(
    base_augmentation_layer_3d.BaseAugmentationLayer3D
):
    """A preprocessing layer which randomly drops point within a randomly
    generated frustum during training.

    This layer will randomly select a point from the point cloud as the center
    of a frustum then generate a frustum based on r_distance, theta_width, and
    phi_width. Points inside the selected frustum are randomly dropped
    (setting all features to zero) based on drop_rate. The point_clouds tensor
    shape must be specific and cannot be dynamic.

    Input shape:
      point_clouds: 3D (multi frames) float32 Tensor with shape
        [num of frames, num of points, num of point features].
        The first 5 features are [x, y, z, class, range].
      bounding_boxes: 3D (multi frames) float32 Tensor with shape
        [num of frames, num of boxes, num of box features].
        The first 7 features are [x, y, z, dx, dy, dz, phi].

    Output shape:
      A dictionary of Tensors with the same shape as input Tensors.

    Arguments:
      r_distance: A float scalar sets the starting distance of a frustum.
      theta_width: A float scalar sets the theta width of a frustum.
      phi_width: A float scalar sets the phi width of a frustum.
      drop_rate: A float scalar sets the probability threshold for dropping the
        points.
      exclude_classes: An optional int scalar or a list of ints. Points with the
        specified class(es) will not be dropped.

    """

    def __init__(
        self,
        r_distance,
        theta_width,
        phi_width,
        drop_rate=None,
        exclude_classes=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if not isinstance(exclude_classes, (tuple, list)):
            exclude_classes = [exclude_classes]

        if r_distance < 0:
            raise ValueError(
                f"r_distance must be >=0, but got r_distance={r_distance}"
            )
        if theta_width < 0:
            raise ValueError(
                f"theta_width must be >=0, but got theta_width={theta_width}"
            )
        if phi_width < 0:
            raise ValueError(
                f"phi_width must be >=0, but got phi_width={phi_width}"
            )
        drop_rate = drop_rate if drop_rate else 0.0
        if drop_rate > 1:
            raise ValueError(
                f"drop_rate must be <=1, but got drop_rate={drop_rate}"
            )

        self._r_distance = r_distance
        self._theta_width = theta_width
        self._phi_width = phi_width
        keep_probability = 1 - drop_rate
        self._keep_probability = keep_probability
        self._exclude_classes = exclude_classes

    def get_config(self):
        return {
            "r_distance": self._r_distance,
            "theta_width": self._theta_width,
            "phi_width": self._phi_width,
            "drop_rate": 1 - self._keep_probability,
            "exclude_classes": self._exclude_classes,
        }

    def get_random_transformation(self, point_clouds, **kwargs):
        # Randomly select a point from the first frame as the center of the
        # frustum.
        valid_points = point_clouds[0, :, POINTCLOUD_LABEL_INDEX] > 0
        num_valid_points = tf.math.reduce_sum(tf.cast(valid_points, tf.int32))
        randomly_select_point_index = tf.random.uniform(
            (), minval=0, maxval=num_valid_points, dtype=tf.int32
        )
        randomly_select_frustum_center = tf.boolean_mask(
            point_clouds[0], valid_points, axis=0
        )[randomly_select_point_index, :POINTCLOUD_LABEL_INDEX]
        num_frames, num_points, _ = point_clouds.get_shape().as_list()
        frustum_mask = []
        for f in range(num_frames):
            frustum_mask.append(
                point_cloud.within_a_frustum(
                    point_clouds[f],
                    randomly_select_frustum_center,
                    self._r_distance,
                    self._theta_width,
                    self._phi_width,
                )[tf.newaxis, :, tf.newaxis]
            )
        frustum_mask = tf.concat(frustum_mask, axis=0)
        # Generate mask along point dimension.
        random_point_mask = (
            random.uniform(
                [1, num_points, 1],
                minval=0.0,
                maxval=1,
                seed=self._random_generator,
            )
            < self._keep_probability
        )
        # Do not drop points outside the frustum mask.
        random_point_mask = tf.where(~frustum_mask, True, random_point_mask)
        return {"point_mask": random_point_mask}

    def augment_point_clouds_bounding_boxes(
        self, point_clouds, bounding_boxes, transformation, **kwargs
    ):
        point_mask = transformation["point_mask"]

        # Do not drop points that are protected by setting the corresponding
        # point_mask = 1.0.
        protected_points = tf.zeros_like(point_clouds[0, :, -1], dtype=tf.bool)
        for excluded_class in self._exclude_classes:
            protected_points |= point_clouds[0, :, -1] == excluded_class
        point_mask = tf.where(
            protected_points[tf.newaxis, :, tf.newaxis], True, point_mask
        )
        point_clouds = tf.where(point_mask, point_clouds, 0.0)
        return (point_clouds, bounding_boxes)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""All Backbone presets"""

from keras_cv.src.models.backbones.csp_darknet import (
    csp_darknet_backbone_presets,
)
from keras_cv.src.models.backbones.densenet import densenet_backbone_presets
from keras_cv.src.models.backbones.efficientnet_lite import (
    efficientnet_lite_backbone_presets,
)
from keras_cv.src.models.backbones.efficientnet_v1 import (
    efficientnet_v1_backbone_presets,
)
from keras_cv.src.models.backbones.efficientnet_v2 import (
    efficientnet_v2_backbone_presets,
)
from keras_cv.src.models.backbones.mobilenet_v3 import (
    mobilenet_v3_backbone_presets,
)
from keras_cv.src.models.backbones.resnet_v1 import resnet_v1_backbone_presets
from keras_cv.src.models.backbones.resnet_v2 import resnet_v2_backbone_presets
from keras_cv.src.models.backbones.video_swin import video_swin_backbone_presets
from keras_cv.src.models.backbones.vit_det import vit_det_backbone_presets
from keras_cv.src.models.object_detection.yolo_v8 import (
    yolo_v8_backbone_presets,
)

backbone_presets_no_weights = {
    **resnet_v1_backbone_presets.backbone_presets_no_weights,
    **resnet_v2_backbone_presets.backbone_presets_no_weights,
    **mobilenet_v3_backbone_presets.backbone_presets_no_weights,
    **csp_darknet_backbone_presets.backbone_presets_no_weights,
    **efficientnet_v1_backbone_presets.backbone_presets_no_weights,
    **efficientnet_v2_backbone_presets.backbone_presets_no_weights,
    **densenet_backbone_presets.backbone_presets_no_weights,
    **efficientnet_lite_backbone_presets.backbone_presets_no_weights,
    **yolo_v8_backbone_presets.backbone_presets_no_weights,
    **vit_det_backbone_presets.backbone_presets_no_weights,
    **video_swin_backbone_presets.backbone_presets_no_weights,
}

backbone_presets_with_weights = {
    **resnet_v1_backbone_presets.backbone_presets_with_weights,
    **resnet_v2_backbone_presets.backbone_presets_with_weights,
    **mobilenet_v3_backbone_presets.backbone_presets_with_weights,
    **csp_darknet_backbone_presets.backbone_presets_with_weights,
    **efficientnet_v1_backbone_presets.backbone_presets_with_weights,
    **efficientnet_v2_backbone_presets.backbone_presets_with_weights,
    **densenet_backbone_presets.backbone_presets_with_weights,
    **efficientnet_lite_backbone_presets.backbone_presets_with_weights,
    **yolo_v8_backbone_presets.backbone_presets_with_weights,
    **vit_det_backbone_presets.backbone_presets_with_weights,
    **video_swin_backbone_presets.backbone_presets_with_weights,
}

backbone_presets = {
    **backbone_presets_no_weights,
    **backbone_presets_with_weights,
}



# Copyright 2024 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Video Swin model preset configurations."""

backbone_presets_no_weights = {
    "videoswin_tiny": {
        "metadata": {
            "description": ("A tiny Video Swin backbone architecture."),
            "params": 27_850_470,
            "official_name": "VideoSwinT",
            "path": "video_swin",
        },
    },
    "videoswin_small": {
        "metadata": {
            "description": ("A small Video Swin backbone architecture."),
            "params": 49_509_078,
            "official_name": "VideoSwinS",
            "path": "video_swin",
        },
    },
    "videoswin_base": {
        "metadata": {
            "description": ("A base Video Swin backbone architecture."),
            "params": 87_638_984,
            "official_name": "VideoSwinB",
            "path": "video_swin",
        },
    },
}

backbone_presets_with_weights = {
    "videoswin_tiny_kinetics400": {
        "metadata": {
            "description": (
                "A tiny Video Swin backbone architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
            ),
            "params": 27_850_470,
            "official_name": "VideoSwinT",
            "path": "video_swin",
        },
    },
    "videoswin_small_kinetics400": {
        "metadata": {
            "description": (
                "A small Video Swin backbone architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "80.6% top1 and 94.5% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 49_509_078,
            "official_name": "VideoSwinS",
            "path": "video_swin",
        },
    },
    "videoswin_base_kinetics400": {
        "metadata": {
            "description": (
                "A base Video Swin backbone architecture. "
                "It is pretrained on ImageNet 1K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "80.6% top1 and 94.6% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 87_638_984,
            "official_name": "VideoSwinB",
            "path": "video_swin",
        },
    },
    "videoswin_base_kinetics400_imagenet22k": {
        "metadata": {
            "description": (
                "A base Video Swin backbone architecture. "
                "It is pretrained on ImageNet 22K dataset, and "
                "trained on Kinetics 400 dataset. "
                "Published weight is capable of scoring "
                "82.7% top1 and 95.5% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 87_638_984,
            "official_name": "VideoSwinB",
            "path": "video_swin",
        },
    },
    "videoswin_base_kinetics600_imagenet22k": {
        "metadata": {
            "description": (
                "A base Video Swin backbone architecture. "
                "It is pretrained on ImageNet 22K dataset, and "
                "trained on Kinetics 600 dataset. "
                "Published weight is capable of scoring "
                "84.0% top1 and 96.5% top5 accuracy on the "
                "Kinetics 600 dataset"
            ),
            "params": 87_638_984,
            "official_name": "VideoSwinB",
            "path": "video_swin",
        },
    },
    "videoswin_base_something_something_v2": {
        "metadata": {
            "description": (
                "A base Video Swin backbone architecture. "
                "It is pretrained on Kinetics 400 dataset, and "
                "trained on Something Something V2 dataset. "
                "Published weight is capable of scoring "
                "69.6% top1 and 92.7% top5 accuracy on the "
                "Kinetics 400 dataset"
            ),
            "params": 87_638_984,
            "official_name": "VideoSwinB",
            "path": "video_swin",
        },
    },
}

backbone_presets = {
    **backbone_presets_no_weights,
    **backbone_presets_with_weights,
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models.feature_extractor.clip.clip_encoder import CLIPEncoder


@keras_cv_export("keras_cv.models.feature_extractor.CLIPTextEncoder")
class CLIPTextEncoder(keras.Model):
    def __init__(
        self,
        transformer_width,
        transformer_layers,
        transformer_heads,
        vocab_size,
        embed_dim,
        context_length,
        **kwargs,
    ):
        super().__init__(
            **kwargs,
        )
        self.transformer_width = transformer_width
        self.transformer_layers = transformer_layers
        self.transformer_heads = transformer_heads
        self.vocab_size = vocab_size
        self.embed_dim = embed_dim
        self.context_length = context_length
        self.token_embedding = keras.layers.Embedding(
            vocab_size,
            transformer_width,
            name="token_embedding",
        )
        self.positional_embedding = keras.layers.Embedding(
            self.context_length,
            transformer_width,
            name="positional_embedding",
        )

        self.encoder = CLIPEncoder(
            width=transformer_width,
            num_layers=transformer_layers,
            heads=transformer_heads,
            name="clip_encoder",
        )
        self.ln_final = keras.layers.LayerNormalization(name="ln_final")

        self.text_projector = keras.layers.Dense(
            embed_dim, name="text_projector", use_bias=False
        )

    def build(self, input_shape):
        self.token_embedding.build(input_shape)
        self.positional_embedding.build([1, self.context_length])
        self.encoder.build(None)
        self.ln_final.build([None, None, self.transformer_width])
        self.text_projector.build([None, self.transformer_width])
        self.built = True

    def compute_output_shape(self, input_shape):
        return [input_shape[0], self.embed_dim]

    def call(self, inputs, attention_mask=None):
        token_embedding = self.token_embedding(inputs)
        position_ids = ops.expand_dims(
            ops.arange(self.context_length, dtype="int32"), 0
        )
        position_embedding = self.positional_embedding(position_ids)
        position_embedding = ops.tile(
            position_embedding, repeats=(ops.shape(inputs)[0], 1, 1)
        )
        causal_attention_mask = ops.ones(
            (self.context_length, self.context_length)
        )
        # Zero out the lower diagonal
        causal_attention_mask = ops.triu(causal_attention_mask)
        causal_attention_mask = ops.cast(causal_attention_mask, "float32")
        attention_mask = ops.cast(attention_mask, dtype="float32")
        expanded_mask = ops.tile(
            attention_mask[:, None, None, :], (1, 1, self.context_length, 1)
        )
        expanded_mask = (1.0 - expanded_mask) * (-1e8)
        encoded_output = self.encoder(
            token_embedding + position_embedding,
            causal_attention_mask=causal_attention_mask,
            attention_mask=expanded_mask,
        )
        layer_norm = self.ln_final(encoded_output)
        indices = ops.expand_dims(
            ops.cast(ops.argmax(inputs, axis=-1), "int32"), axis=-1
        )
        selected_features = ops.take_along_axis(
            layer_norm, indices[:, :, None], axis=1
        )
        text_features = self.text_projector(selected_features)
        output = ops.squeeze(text_features, axis=1)
        return output

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "transformer_width": self.transformer_width,
                "transformer_layers": self.transformer_layers,
                "transformer_heads": self.transformer_heads,
                "vocab_size": self.vocab_size,
                "embed_dim": self.embed_dim,
                "context_length": self.context_length,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import tensorflow_datasets as tfds

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export


def curry_map_function(bounding_box_format):
    """Mapping function to create batched image and bbox coordinates"""

    def apply(inputs):
        images = inputs["image"]
        bounding_boxes = inputs["objects"]["bbox"]
        labels = inputs["objects"]["label"]
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            images=images,
            source="rel_yxyx",
            target=bounding_box_format,
        )

        bounding_boxes = {"boxes": bounding_boxes, "classes": labels}

        outputs = {"images": images, "bounding_boxes": bounding_boxes}
        return outputs

    return apply


@keras_cv_export("keras_cv.datasets.pascal_voc.load")
def load(
    split,
    bounding_box_format,
    batch_size=None,
    shuffle_files=True,
    shuffle_buffer=None,
    dataset="voc/2007",
):
    """Loads the PascalVOC 2007 dataset.

    Example:
    ```python
    dataset, ds_info = keras_cv.datasets.pascal_voc.load(
        split="train", bounding_box_format="xywh", batch_size=9
    )
    ```

    Args:
        split: the split string passed to the `tensorflow_datasets.load()` call.
            Should be one of "train", "test", or "validation."
        bounding_box_format: the keras_cv bounding box format to load the boxes
            into. For a list of supported formats, please refer
            [to the keras.io docs](https://keras.io/api/keras_cv/bounding_box/formats/)
            for more details on supported bounding box formats.
        batch_size: how many instances to include in batches after loading
        shuffle_buffer: the size of the buffer to use in shuffling.
        shuffle_files: (Optional) whether to shuffle files, defaults to
            True.
        dataset: (Optional) the PascalVOC dataset to load from. Should be either
            'voc/2007' or 'voc/2012', defaults to 'voc/2007'.

    Returns:
        tf.data.Dataset containing PascalVOC. Each entry is a dictionary
        containing keys {"images": images, "bounding_boxes": bounding_boxes}
        where images is a Tensor of shape [batch, H, W, 3] and bounding_boxes is
        a `tf.RaggedTensor` of shape [batch, None, 5].
    """  # noqa: E501
    if dataset not in ["voc/2007", "voc/2012"]:
        raise ValueError(
            "keras_cv.datasets.pascal_voc.load() expects the `dataset` "
            "argument to be either 'voc/2007' or 'voc/2012', but got "
            f"`dataset={dataset}`."
        )
    dataset, dataset_info = tfds.load(
        dataset, split=split, shuffle_files=shuffle_files, with_info=True
    )
    dataset = dataset.map(
        curry_map_function(bounding_box_format=bounding_box_format),
        num_parallel_calls=tf.data.AUTOTUNE,
    )

    if shuffle_buffer:
        dataset = dataset.shuffle(shuffle_buffer, reshuffle_each_iteration=True)

    if batch_size is not None:
        dataset = dataset.apply(
            tf.data.experimental.dense_to_ragged_batch(batch_size=batch_size)
        )
    return dataset, dataset_info



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)


@keras_cv_export("keras_cv.layers.Grayscale")
class Grayscale(VectorizedBaseImageAugmentationLayer):
    """Grayscale is a preprocessing layer that transforms RGB images to
    Grayscale images.
    Input images should have values in the range of [0, 255].

    Input shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format
    Output shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format
    Args:
        output_channels.
            Number color channels present in the output image.
            The output_channels can be 1 or 3. RGB image with shape
            (..., height, width, 3) will have the following shapes
            after the `Grayscale` operation:
                 a. (..., height, width, 1) if output_channels = 1
                 b. (..., height, width, 3) if output_channels = 3.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    to_grayscale = keras_cv.layers.preprocessing.Grayscale()
    augmented_images = to_grayscale(images)
    ```
    """

    def __init__(self, output_channels=1, **kwargs):
        super().__init__(**kwargs)
        self.output_channels = output_channels
        self._check_input_params(output_channels)

    def _check_input_params(self, output_channels):
        if output_channels not in [1, 3]:
            raise ValueError(
                "Received invalid argument output_channels. "
                f"output_channels must be in 1 or 3. Got {output_channels}"
            )
        self.output_channels = output_channels

    def compute_ragged_image_signature(self, images):
        ragged_spec = tf.RaggedTensorSpec(
            shape=images.shape[1:3] + (self.output_channels,),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations=None, **kwargs):
        grayscale = tf.image.rgb_to_grayscale(images)
        if self.output_channels == 1:
            return grayscale
        elif self.output_channels == 3:
            return tf.image.grayscale_to_rgb(grayscale)
        else:
            raise ValueError("Unsupported value for `output_channels`.")

    def augment_bounding_boxes(self, bounding_boxes, **kwargs):
        return bounding_boxes

    def augment_labels(self, labels, transformations=None, **kwargs):
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def get_config(self):
        config = {
            "output_channels": self.output_channels,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types

from tensorflow import keras  # noqa: F403, F401
from tensorflow.keras import *  # noqa: F403, F401

from keras_cv.src.backend import config  # noqa: F403, F401

_KERAS_CORE_ALIASES = {
    "utils->saving": [
        "register_keras_serializable",
        "deserialize_keras_object",
        "serialize_keras_object",
        "get_registered_object",
    ],
    "models->saving": ["load_model"],
}

if not hasattr(keras, "saving"):
    keras.saving = types.SimpleNamespace()

# add aliases
for key, value in _KERAS_CORE_ALIASES.items():
    src, _, dst = key.partition("->")
    src = src.split(".")
    dst = dst.split(".")

    src_mod, dst_mod = keras, keras

    # navigate to where we want to alias the attributes
    for mod in src:
        src_mod = getattr(src_mod, mod)
    for mod in dst:
        dst_mod = getattr(dst_mod, mod)

    # add an alias for each attribute
    for attr in value:
        if isinstance(attr, tuple):
            src_attr, dst_attr = attr
        else:
            src_attr, dst_attr = attr, attr
        attr_val = getattr(src_mod, src_attr)
        setattr(dst_mod, dst_attr, attr_val)

# TF Keras doesn't have this rename.
keras.activations.silu = keras.activations.swish



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import keras
import tensorflow as tf
import tree

if hasattr(keras, "src"):
    keras_backend = keras.src.backend
else:
    keras_backend = keras.backend

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import config
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.backend import scope
from keras_cv.src.utils import preprocessing

# In order to support both unbatched and batched inputs, the horizontal
# and vertical axis is reverse indexed
H_AXIS = -3
W_AXIS = -2

IMAGES = "images"
LABELS = "labels"
TARGETS = "targets"
BOUNDING_BOXES = "bounding_boxes"
KEYPOINTS = "keypoints"
SEGMENTATION_MASKS = "segmentation_masks"
IS_DICT = "is_dict"
USE_TARGETS = "use_targets"


@keras_cv_export("keras_cv.layers.BaseImageAugmentationLayer")
class BaseImageAugmentationLayer(keras.layers.Layer):
    """Abstract base layer for image augmentation.

    This layer contains base functionalities for preprocessing layers which
    augment image related data, e.g. image and in the future, label and bounding
    boxes. The subclasses could avoid making certain mistakes and reduce code
    duplications.

    This layer requires you to implement one method: `augment_image()`, which
    augments one single image during the training. There are a few additional
    methods that you can implement for added functionality on the layer:

    `augment_label()`, which handles label augmentation if the layer supports
    that.

    `augment_bounding_boxes()`, which handles the bounding box augmentation, if
    the layer supports that.

    `get_random_transformation()`, which should produce a random transformation
    setting. The transformation object, which could be of any type, will be
    passed to `augment_image`, `augment_label` and `augment_bounding_boxes`, to
    coordinate the randomness behaviour, e.g., in the RandomFlip layer, the
    image and bounding_boxes should be changed in the same way.

    The `call()` method supports two formats of inputs:
    1. A single image tensor with shape (height, width, channels) or
       (batch_size, height, width, channels)
    2. A dict of tensors with any of the following keys (note that `"images"`
       must be present):
       * `"images"` - Image Tensor with shape (height, width, channels) or
        (batch_size, height, width, channels)
       * `"labels"` - One-hot encoded classification labels Tensor with shape
        (num_classes) or (batch_size, num_classes)
       * `"bounding_boxes"` - A dictionary with keys:
         * `"boxes"` - Tensor with shape (num_boxes, 4) or (batch_size,
          num_boxes, 4)
         * `"classes"` - Tensor of class labels for boxes with shape (num_boxes,
          num_classes) or (batch_size, num_boxes, num_classes).
       Any other keys included in this dictionary will be ignored and unmodified
       by an augmentation layer.

    The output of the `call()` will be the same structure as the inputs.

    The `call()` will unpack the inputs, forward to the correct function, and
    pack the output back to the same structure as the inputs.

    By default, the `call()` method leverages the `tf.vectorized_map()`
    function. Auto-vectorization can be disabled by setting
    `self.auto_vectorize = False` in your `__init__()` method. When disabled,
    `call()` instead relies on `tf.map_fn()`. For example:

    ```python
    class SubclassLayer(keras_cv.BaseImageAugmentationLayer):
      def __init__(self):
        super().__init__()
        self.auto_vectorize = False
    ```

    Example:

    ```python
    class RandomContrast(keras_cv.BaseImageAugmentationLayer):

      def __init__(self, factor=(0.5, 1.5), **kwargs):
        super().__init__(**kwargs)
        self._factor = factor

      def augment_image(self, image, transformation):
        random_factor = tf.random.uniform([], self._factor[0], self._factor[1])
        mean = tf.math.reduced_mean(inputs, axis=-1, keep_dim=True)
        return (inputs - mean) * random_factor + mean
    ```

    Note that since the randomness is also a common functionality, this layer
    also includes a keras_backend.RandomGenerator, which can be used to
    produce the random numbers. The random number generator is stored in the
    `self._random_generator` attribute.
    """

    def __init__(self, seed=None, **kwargs):
        if seed is not None:
            self._random_generator = tf.random.Generator.from_seed(seed=seed)
        else:
            self._random_generator = tf.random.get_global_generator()
        super().__init__(**kwargs)

        self._allow_non_tensor_positional_args = True
        self.built = True
        self._convert_input_args = False

    @property
    def force_output_ragged_images(self):
        """Control whether to force outputting of ragged images."""
        return getattr(self, "_force_output_ragged_images", False)

    @force_output_ragged_images.setter
    def force_output_ragged_images(self, force_output_ragged_images):
        self._force_output_ragged_images = force_output_ragged_images

    @property
    def force_output_dense_images(self):
        """Control whether to force outputting of dense images."""
        return getattr(self, "_force_output_dense_images", False)

    @force_output_dense_images.setter
    def force_output_dense_images(self, force_output_dense_images):
        self._force_output_dense_images = force_output_dense_images

    @property
    def auto_vectorize(self):
        """Control whether automatic vectorization occurs.

        By default, the `call()` method leverages the `tf.vectorized_map()`
        function. Auto-vectorization can be disabled by setting
        `self.auto_vectorize = False` in your `__init__()` method. When
        disabled, `call()` instead relies on `tf.map_fn()`. For example:

        ```python
        class SubclassLayer(BaseImageAugmentationLayer):
          def __init__(self):
            super().__init__()
            self.auto_vectorize = False
        ```
        """
        return getattr(self, "_auto_vectorize", True)

    @auto_vectorize.setter
    def auto_vectorize(self, auto_vectorize):
        self._auto_vectorize = auto_vectorize

    def compute_image_signature(self, images):
        """Computes the output image signature for the `augment_image()`
        function.

        Must be overridden to return tensors with different shapes than the
        input images. By default, returns either a `tf.RaggedTensorSpec`
        matching the input image spec, or a `tf.TensorSpec` matching the input
        image spec.
        """
        if self.force_output_dense_images:
            return tf.TensorSpec(images.shape[1:], self.compute_dtype)
        if self.force_output_ragged_images or isinstance(
            images, tf.RaggedTensor
        ):
            ragged_spec = tf.RaggedTensorSpec(
                shape=images.shape[1:],
                ragged_rank=1,
                dtype=self.compute_dtype,
            )
            return ragged_spec
        return tf.TensorSpec(images.shape[1:], self.compute_dtype)

    # TODO(lukewood): promote to user facing API if needed
    def _compute_bounding_box_signature(self, bounding_boxes):
        return {
            "boxes": tf.RaggedTensorSpec(
                shape=[None, 4],
                ragged_rank=1,
                dtype=self.compute_dtype,
            ),
            "classes": tf.RaggedTensorSpec(
                shape=[None], dtype=self.compute_dtype
            ),
        }

    # TODO(lukewood): promote to user facing API if needed
    def _compute_keypoints_signature(self, keypoints):
        if isinstance(keypoints, tf.RaggedTensor):
            ragged_spec = tf.RaggedTensorSpec(
                shape=keypoints.shape[1:],
                ragged_rank=1,
                dtype=self.compute_dtype,
            )
            return ragged_spec

        return tf.TensorSpec(
            shape=keypoints.shape[1:],
            dtype=self.compute_dtype,
        )

    # TODO(lukewood): promote to user facing API if needed
    def _compute_target_signature(self, targets):
        return tf.TensorSpec(targets.shape[1:], self.compute_dtype)

    def _compute_output_signature(self, inputs):
        fn_output_signature = {
            IMAGES: self.compute_image_signature(inputs[IMAGES])
        }
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)

        if bounding_boxes is not None:
            fn_output_signature[BOUNDING_BOXES] = (
                self._compute_bounding_box_signature(bounding_boxes)
            )

        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        if segmentation_masks is not None:
            fn_output_signature[SEGMENTATION_MASKS] = (
                self.compute_image_signature(segmentation_masks)
            )

        keypoints = inputs.get(KEYPOINTS, None)
        if keypoints is not None:
            fn_output_signature[KEYPOINTS] = self._compute_keypoints_signature(
                keypoints
            )

        labels = inputs.get(LABELS, None)
        if labels is not None:
            fn_output_signature[LABELS] = self._compute_target_signature(labels)

        return fn_output_signature

    @staticmethod
    def _any_ragged(inputs):
        if isinstance(inputs[IMAGES], tf.RaggedTensor):
            return True
        if BOUNDING_BOXES in inputs:
            return True
        if KEYPOINTS in inputs:
            return True
        return False

    def _map_fn(self, func, inputs):
        """Returns either tf.map_fn or tf.vectorized_map based on the provided
        inputs.

        Args:
            inputs: dictionary of inputs provided to map_fn.
        """
        if self._any_ragged(inputs) or self.force_output_ragged_images:
            return tf.map_fn(
                func,
                inputs,
                fn_output_signature=self._compute_output_signature(inputs),
            )
        if self.auto_vectorize:
            return tf.vectorized_map(func, inputs)
        return tf.map_fn(func, inputs)

    def augment_image(self, image, transformation, **kwargs):
        """Augment a single image during training.

        Args:
          image: 3D image input tensor to the layer. Forwarded from
            `layer.call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_label(self, label, transformation, **kwargs):
        """Augment a single label during training.

        Args:
          label: 1D label to the layer. Forwarded from `layer.call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 1D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_target(self, target, transformation, **kwargs):
        """Augment a single target during training.

        Args:
          target: 1D label to the layer. Forwarded from `layer.call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 1D tensor, which will be forward to `layer.call()`.
        """
        return self.augment_label(target, transformation)

    def augment_bounding_boxes(self, bounding_boxes, transformation, **kwargs):
        """Augment bounding boxes for one image during training.

        Args:
          bounding_boxes: 2D bounding boxes to the layer. Forwarded from
            `call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 2D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_keypoints(self, keypoints, transformation, **kwargs):
        """Augment keypoints for one image during training.

        Args:
          keypoints: 2D keypoints input tensor to the layer. Forwarded from
            `layer.call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 2D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        """Augment a single image's segmentation mask during training.

        Args:
          segmentation_mask: 3D segmentation mask input tensor to the layer.
            This should generally have the shape [H, W, 1], or in some cases
            [H, W, C] for multilabeled data. Forwarded from `layer.call()`.
          transformation: The transformation object produced by
            `get_random_transformation`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 3D tensor containing the augmented segmentation mask, which
          will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def get_random_transformation(
        self,
        image=None,
        label=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_mask=None,
    ):
        """Produce random transformation config for one single input.

        This is used to produce same randomness between
        image/label/bounding_box.

        Args:
          image: 3D image tensor from inputs.
          label: optional 1D label tensor from inputs.
          bounding_boxes: optional 2D bounding boxes tensor from inputs.
          segmentation_mask: optional 3D segmentation mask tensor from inputs.

        Returns:
          Any type of object, which will be forwarded to `augment_image`,
          `augment_label` and `augment_bounding_box` as the `transformation`
          parameter.
        """
        return None

    def call(self, inputs):
        # try to convert a given backend native tensor to TensorFlow tensor
        # before passing it over to TFDataScope
        is_tf_backend = config.backend() == "tensorflow"
        is_in_tf_graph = not tf.executing_eagerly()
        contains_ragged = lambda y: any(
            tree.map_structure(
                lambda x: isinstance(x, (tf.RaggedTensor, tf.SparseTensor)),
                tree.flatten(y),
            )
        )
        inputs_contain_ragged = contains_ragged(inputs)
        if not is_tf_backend and not inputs_contain_ragged:
            inputs = tree.map_structure(
                lambda x: tf.convert_to_tensor(x), inputs
            )
        with scope.TFDataScope():
            inputs = self._ensure_inputs_are_compute_dtype(inputs)
            inputs, metadata = self._format_inputs(inputs)
            images = inputs[IMAGES]
            if images.shape.rank == 3:
                outputs = self._format_output(self._augment(inputs), metadata)
            elif images.shape.rank == 4:
                outputs = self._format_output(
                    self._batch_augment(inputs), metadata
                )
            else:
                raise ValueError(
                    "Image augmentation layers are expecting inputs to be "
                    "rank 3 (HWC) or 4D (NHWC) tensors. Got shape: "
                    f"{images.shape}"
                )
        # convert the outputs to backend native tensors if none of them
        # contain RaggedTensors. Note that if the user passed in Raggeds
        # but the outputs are dense, we still don't want to convert to
        # backend native tensors. This is to avoid breaking TF data
        # pipelines that can't easily be ported to become backend
        # agnostic.
        # Skip this step for TF backend or if in `tf.graph` like `tf.data`.
        if not is_tf_backend and not is_in_tf_graph:
            if not inputs_contain_ragged and not contains_ragged(outputs):
                outputs = tree.map_structure(
                    # some layers return None, handle that case when
                    # converting to tensors
                    lambda x: ops.convert_to_tensor(x) if x is not None else x,
                    outputs,
                )
        return outputs

    def _augment(self, inputs):
        raw_image = inputs.get(IMAGES, None)
        image = raw_image
        label = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        keypoints = inputs.get(KEYPOINTS, None)
        segmentation_mask = inputs.get(SEGMENTATION_MASKS, None)

        image_ragged = isinstance(image, tf.RaggedTensor)
        # At this point, the tensor is not actually ragged as we have mapped
        # over the batch axis. This call is required to make `tf.shape()` behave
        # as users subclassing the layer expect.
        if image_ragged:
            image = image.to_tensor()

        transformation = self.get_random_transformation(
            image=image,
            label=label,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_mask=segmentation_mask,
        )

        image = self.augment_image(
            image,
            transformation=transformation,
            bounding_boxes=bounding_boxes,
            label=label,
        )

        if (
            image_ragged and not self.force_output_dense_images
        ) or self.force_output_ragged_images:
            image = tf.RaggedTensor.from_tensor(image)

        result = {IMAGES: image}
        if label is not None:
            label = self.augment_target(
                label,
                transformation=transformation,
                bounding_boxes=bounding_boxes,
                image=image,
            )
            result[LABELS] = label

        if bounding_boxes is not None:
            bounding_boxes = bounding_box.to_dense(bounding_boxes)

            bounding_boxes = self.augment_bounding_boxes(
                bounding_boxes,
                transformation=transformation,
                label=label,
                image=raw_image,
            )

            bounding_boxes = bounding_box.to_ragged(
                bounding_boxes, dtype=self.compute_dtype
            )
            result[BOUNDING_BOXES] = bounding_boxes

        if keypoints is not None:
            keypoints = self.augment_keypoints(
                keypoints,
                transformation=transformation,
                label=label,
                bounding_boxes=bounding_boxes,
                image=image,
            )
            result[KEYPOINTS] = keypoints
        if segmentation_mask is not None:
            segmentation_mask = self.augment_segmentation_mask(
                segmentation_mask,
                transformation=transformation,
            )
            result[SEGMENTATION_MASKS] = segmentation_mask

        # preserve any additional inputs unmodified by this layer.
        for key in inputs.keys() - result.keys():
            result[key] = inputs[key]
        return result

    def _batch_augment(self, inputs):
        return self._map_fn(self._augment, inputs)

    def _format_inputs(self, inputs):
        metadata = {IS_DICT: True, USE_TARGETS: False}
        if tf.is_tensor(inputs):
            # single image input tensor
            metadata[IS_DICT] = False
            inputs = {IMAGES: inputs}
            return inputs, metadata

        if not isinstance(inputs, dict):
            raise ValueError(
                "Expect the inputs to be image tensor or dict. Got "
                f"inputs={inputs} of type {type(inputs)}"
            )

        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES] = self._format_bounding_boxes(
                inputs[BOUNDING_BOXES]
            )

        if isinstance(inputs, dict) and TARGETS in inputs:
            # TODO(scottzhu): Check if it only contains the valid keys
            inputs[LABELS] = inputs[TARGETS]
            del inputs[TARGETS]
            metadata[USE_TARGETS] = True
            return inputs, metadata

        return inputs, metadata

    def _format_bounding_boxes(self, bounding_boxes):
        # We can't catch the case where this is None, sometimes RaggedTensor
        # drops this dimension
        if "classes" not in bounding_boxes:
            raise ValueError(
                "Bounding boxes are missing class_id. If you would like to pad "
                "the bounding boxes with class_id, use: "
                "`bounding_boxes['classes'] = "
                "tf.ones_like(bounding_boxes['boxes'])`."
            )
        return bounding_boxes

    def _format_output(self, output, metadata):
        if not metadata[IS_DICT]:
            return output[IMAGES]
        elif metadata[USE_TARGETS]:
            output[TARGETS] = output[LABELS]
            del output[LABELS]
        return output

    def _ensure_inputs_are_compute_dtype(self, inputs):
        if not isinstance(inputs, dict):
            return preprocessing.ensure_tensor(
                inputs,
                self.compute_dtype,
            )
        # Copy the input dict before we mutate it.
        inputs = dict(inputs)
        inputs[IMAGES] = preprocessing.ensure_tensor(
            inputs[IMAGES],
            self.compute_dtype,
        )
        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES] = bounding_box.ensure_tensor(
                inputs[BOUNDING_BOXES], dtype=self.compute_dtype
            )
        return inputs



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.layers.vit_det_layers import MLP


@keras_cv_export(
    "keras_cv.layers.MultiHeadAttentionWithDownsampling",
    package="keras_cv.layers",
)
class MultiHeadAttentionWithDownsampling(keras.layers.Layer):
    """Multi-Head Attention with downsampling.

    An attention layer that allows for downscaling the size of the embedding
    after projection to queries, keys, and values.

    This layer first downscales the features of input queries, keys, and
    values using a dense layer. Multi-head attention is then performed
    and the attention map is projected back (upscaled) to the number of
    input features.

    Args:
        num_heads (int): Number of attention heads.
        key_dim (int): Size of each attention head for query, key, and
            value.
        downsample_rate (int, optional): The factor by which to downscale the
            input features i.e. the input features of size `key_dim` are
            projected down to `key_dim // downsample_rate`.

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)
    """  # noqa: E501

    def __init__(self, num_heads, key_dim, downsample_rate=1, **kwargs):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.downsample_rate = downsample_rate
        self.internal_dims = key_dim // downsample_rate

        # Downsample
        self.query_proj = keras.layers.Dense(
            self.internal_dims * self.num_heads
        )
        self.key_proj = keras.layers.Dense(self.internal_dims * self.num_heads)
        self.value_proj = keras.layers.Dense(
            self.internal_dims * self.num_heads
        )

        # Upsample
        self.out_proj = keras.layers.Dense(self.key_dim * self.num_heads)

    def build(self, input_shape=None):
        self.query_proj.build([None, None, self.num_heads * self.key_dim])
        self.key_proj.build([None, None, self.num_heads * self.key_dim])
        self.value_proj.build([None, None, self.num_heads * self.key_dim])
        self.out_proj.build([None, None, self.internal_dims * self.num_heads])
        self.built = True

    def __separate_heads(self, x):
        shape = ops.shape(x)
        B, N, C = shape[0], shape[1], shape[2]
        x = ops.reshape(x, (B, N, self.num_heads, C // self.num_heads))
        return ops.transpose(x, axes=(0, 2, 1, 3))

    def __recombine_heads(self, x):
        shape = ops.shape(x)
        B, N_H, N_T, C_PH = shape[0], shape[1], shape[2], shape[3]
        x = ops.transpose(x, axes=(0, 2, 1, 3))
        return ops.reshape(x, (B, N_T, N_H * C_PH))

    def call(self, query, value, key):
        query = self.query_proj(query)
        key = self.key_proj(key)
        value = self.value_proj(value)

        # Separate into heads
        query = self.__separate_heads(query)
        key = self.__separate_heads(key)
        value = self.__separate_heads(value)

        # Attention
        C_PH = ops.shape(query)[-1]
        out = query @ ops.transpose(key, (0, 1, 3, 2))
        out = out / ops.sqrt(ops.cast(C_PH, dtype=self.compute_dtype))
        out = ops.softmax(out, axis=-1)

        # Get output
        attention_map = out @ value
        attention_map = self.__recombine_heads(attention_map)
        return self.out_proj(attention_map)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_heads": self.num_heads,
                "key_dim": self.key_dim,
                "downsample_rate": self.downsample_rate,
            }
        )
        return config


@keras_cv_export(
    "keras_cv.layers.TwoWayMultiHeadAttention", package="keras_cv.layers"
)
class TwoWayMultiHeadAttention(keras.layers.Layer):
    """Two-way multi-head attention layer.

    Args:
        num_heads (int): Number of attention heads.
        key_dim (int): Size of each attention head for query, key, and
            value.
        mlp_dim (int): Number of hidden dims to use in the mlp block.
        skip_first_layer_pe (bool): A boolean indicating whether to skip the
            first layer positional embeddings.
        attention_downsample_rate (int, optional): The downsample rate to use
            in the attention layers. Defaults to 2.
        activation (str, optional): The activation for the mlp block's output
            layer. Defaults to "relu".

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)
    """  # noqa: E501

    def __init__(
        self,
        num_heads,
        key_dim,
        mlp_dim,
        skip_first_layer_pe,
        attention_downsample_rate=2,
        activation="relu",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.mlp_dim = mlp_dim
        self.skip_first_layer_pe = skip_first_layer_pe
        self.attention_downsample_rate = attention_downsample_rate
        self.activation = activation

        self.self_attention = MultiHeadAttentionWithDownsampling(
            num_heads=num_heads, key_dim=key_dim
        )
        self.layer_norm1 = keras.layers.LayerNormalization(epsilon=1e-5)
        self.cross_attention_token_to_image = (
            MultiHeadAttentionWithDownsampling(
                num_heads=num_heads,
                key_dim=key_dim,
                downsample_rate=attention_downsample_rate,
            )
        )
        self.layer_norm2 = keras.layers.LayerNormalization(epsilon=1e-5)

        self.mlp_block = MLP(
            mlp_dim,
            key_dim * num_heads,
            num_layers=2,
            activation=activation,
        )

        self.layer_norm3 = keras.layers.LayerNormalization(epsilon=1e-5)
        self.cross_attention_image_to_token = (
            MultiHeadAttentionWithDownsampling(
                num_heads=num_heads,
                key_dim=key_dim,
                downsample_rate=attention_downsample_rate,
            )
        )
        self.layer_norm4 = keras.layers.LayerNormalization(epsilon=1e-5)

    def build(self, input_shape=None):
        self.self_attention.build()
        self.layer_norm1.build([None, None, self.num_heads * self.key_dim])
        self.cross_attention_token_to_image.build()
        self.layer_norm2.build([None, None, self.num_heads * self.key_dim])
        self.mlp_block.build([None, None, self.num_heads * self.key_dim])
        self.layer_norm3.build([None, None, self.num_heads * self.key_dim])
        self.cross_attention_image_to_token.build()
        self.layer_norm4.build([None, None, self.num_heads * self.key_dim])
        self.built = True

    def call(self, queries, keys, query_pe, key_pe):
        if self.skip_first_layer_pe:
            queries = self.self_attention(
                query=queries, value=queries, key=queries
            )
        else:
            queries_with_pe = queries + query_pe
            attention_map = self.self_attention(
                query=queries_with_pe, key=queries_with_pe, value=queries
            )
            queries = queries + attention_map
        queries = self.layer_norm1(queries)

        queries_with_pe = queries + query_pe
        keys_with_pe = keys + key_pe
        attention_map = self.cross_attention_token_to_image(
            query=queries_with_pe, key=keys_with_pe, value=keys
        )
        queries = queries + attention_map
        queries = self.layer_norm2(queries)

        mlp_out = self.mlp_block(queries)
        queries = queries + mlp_out
        queries = self.layer_norm3(queries)

        queries_with_pe = queries + query_pe
        keys_with_pe = keys + key_pe
        attention_map = self.cross_attention_image_to_token(
            query=keys_with_pe, key=queries_with_pe, value=queries
        )
        keys = keys + attention_map
        keys = self.layer_norm4(keys)

        return queries, keys

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_heads": self.num_heads,
                "key_dim": self.key_dim,
                "mlp_dim": self.mlp_dim,
                "skip_first_layer_pe": self.skip_first_layer_pe,
                "attention_downsample_rate": self.attention_downsample_rate,
                "activation": self.activation,
            }
        )
        return config


@keras_cv_export(
    "keras_cv.layers.RandomFrequencyPositionalEmbeddings",
    package="keras_cv.layers",
)
class RandomFrequencyPositionalEmbeddings(keras.layers.Layer):
    """Positional encoding using random spatial frequencies.

    This layer maps coordinates/points in 2D space to positional
    encodings using random spatial frequencies.

    Args:
        num_positional_features (int): Number of positional features
            in the output.
        scale (float): The standard deviation of the random frequencies.

    References:
        - [Segment Anything paper](https://arxiv.org/abs/2304.02643)
        - [Segment Anything GitHub](https://github.com/facebookresearch/segment-anything)
    """  # noqa: E501

    def __init__(self, num_positional_features, scale, **kwargs):
        super().__init__(**kwargs)
        self.num_positional_features = num_positional_features
        self.scale = scale
        self.positional_encoding_gaussian_matrix = self.add_weight(
            name="positional_encoding_gaussian_matrix",
            shape=(2, self.num_positional_features),
            dtype=self.variable_dtype,
            trainable=False,
            initializer=keras.initializers.get("normal"),
        )

    def build(self, input_shape=None):
        self.built = True

    def __positional_encodings(self, coords):
        coords = coords * 2 - 1
        coords = coords @ ops.cast(
            self.positional_encoding_gaussian_matrix, dtype=self.compute_dtype
        )
        coords = coords * (2 * math.pi)
        return ops.concatenate([ops.sin(coords), ops.cos(coords)], axis=-1)

    def call(self, size):
        return self.encode_image(size)

    def encode_image(self, size):
        """Generate a positional encoding for an image of any given size.

        Args:
            size (tuple[int, int]): The size of the image.

        Returns:
            tensor: Positional encoding of the image.
        """
        H, W = size
        grid = ops.ones(shape=(H, W), dtype=self.compute_dtype)
        y_embed = ops.cumsum(grid, axis=0) - 0.5
        x_embed = ops.cumsum(grid, axis=1) - 0.5
        y_embed = y_embed / ops.cast(H, self.compute_dtype)
        x_embed = x_embed / ops.cast(W, self.compute_dtype)
        return self.__positional_encodings(
            ops.stack([x_embed, y_embed], axis=-1)
        )

    def encode_coordinates(self, coords_input, image_size):
        """Positionally encode points that are not normalized to `[0, 1]`.

        Args:
            coords_input (tensor): 2D coordinates/points to map.
            image_size (tuple[int, int]): Height and width of the image
                being prompted.

        Returns:
            tensor: Positional encodings of the normalized coordinates.
        """
        coords_normalized = ops.stack(
            [
                coords_input[..., 0] / image_size[1],
                coords_input[..., 1] / image_size[0],
            ],
            axis=-1,
        )
        return self.__positional_encodings(coords_normalized)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_positional_features": self.num_positional_features,
                "scale": self.scale,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export


@keras_cv_export("keras_cv.core.FactorSampler")
class FactorSampler:
    """FactorSampler represents a strength factor for use in an augmentation
    layer.

    FactorSampler should be subclassed and implement a `__call__()` method that
    returns a tf.float32, or a float. This method will be used by preprocessing
    layers to determine the strength of their augmentation. The specific range
    of values supported may vary by layer, but for most layers is the range
    [0, 1].
    """

    def __call__(self, shape=None, dtype="float32"):
        raise NotImplementedError(
            "FactorSampler subclasses must implement a `__call__()` method."
        )

    def get_config(self):
        return {}



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""StableDiffusion Noise scheduler

Adapted from https://github.com/huggingface/diffusers/blob/v0.3.0/src/diffusers/schedulers/scheduling_ddpm.py#L56
"""  # noqa: E501

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import ops
from keras_cv.src.backend import random


@keras_cv_export("keras_cv.models.stable_diffusion.NoiseScheduler")
class NoiseScheduler:
    """
    Args:
        train_timesteps: number of diffusion steps used to train the model.
        beta_start: the starting `beta` value of inference.
        beta_end: the final `beta` value.
        beta_schedule: the beta schedule, a mapping from a beta range to a
            sequence of betas for stepping the model. Choose from `linear` or
            `quadratic`.
        variance_type: options to clip the variance used when adding noise to
            the de-noised sample. Choose from `fixed_small`, `fixed_small_log`,
            `fixed_large`, `fixed_large_log`, `learned` or `learned_range`.
        clip_sample: option to clip predicted sample between -1 and 1 for
            numerical stability.
    """

    def __init__(
        self,
        train_timesteps=1000,
        beta_start=0.0001,
        beta_end=0.02,
        beta_schedule="linear",
        variance_type="fixed_small",
        clip_sample=True,
    ):
        self.train_timesteps = train_timesteps

        if beta_schedule == "linear":
            self.betas = ops.linspace(beta_start, beta_end, train_timesteps)
        elif beta_schedule == "scaled_linear":
            # this schedule is very specific to the latent diffusion model.
            self.betas = (
                ops.linspace(beta_start**0.5, beta_end**0.5, train_timesteps)
                ** 2
            )
        else:
            raise ValueError(f"Invalid beta schedule: {beta_schedule}.")

        self.alphas = 1.0 - self.betas
        self.alphas_cumprod = ops.cumprod(self.alphas)

        self.variance_type = variance_type
        self.clip_sample = clip_sample
        self.seed_generator = random.SeedGenerator(seed=42)

    def _get_variance(self, timestep, predicted_variance=None):
        alpha_prod = self.alphas_cumprod[timestep]
        alpha_prod_prev = (
            self.alphas_cumprod[timestep - 1] if timestep > 0 else 1.0
        )

        variance = (
            (1 - alpha_prod_prev) / (1 - alpha_prod) * self.betas[timestep]
        )

        if self.variance_type == "fixed_small":
            variance = ops.clip(variance, x_min=1e-20, x_max=1)
        elif self.variance_type == "fixed_small_log":
            variance = ops.log(ops.clip(variance, x_min=1e-20, x_max=1))
        elif self.variance_type == "fixed_large":
            variance = self.betas[timestep]
        elif self.variance_type == "fixed_large_log":
            variance = ops.log(self.betas[timestep])
        elif self.variance_type == "learned":
            return predicted_variance
        elif self.variance_type == "learned_range":
            min_log = variance
            max_log = self.betas[timestep]
            frac = (predicted_variance + 1) / 2
            variance = frac * max_log + (1 - frac) * min_log
        else:
            raise ValueError(f"Invalid variance type: {self.variance_type}")

        return variance

    def step(
        self,
        model_output,
        timestep,
        sample,
        predict_epsilon=True,
    ):
        """
        Predict the sample at the previous timestep by reversing the SDE. Core
        function to propagate the diffusion process from the learned model
        outputs (usually the predicted noise).
        Args:
            model_output: a Tensor containing direct output from learned
                diffusion model
            timestep: current discrete timestep in the diffusion chain.
            sample: a Tensor containing the current instance of sample being
                created by diffusion process.
            predict_epsilon: whether the model is predicting noise (epsilon) or
                samples
        Returns:
            The predicted sample at the previous timestep
        """

        if model_output.shape[1] == sample.shape[
            1
        ] * 2 and self.variance_type in [
            "learned",
            "learned_range",
        ]:
            model_output, predicted_variance = ops.split(
                model_output, sample.shape[1], axis=1
            )
        else:
            predicted_variance = None

        # 1. compute alphas, betas
        alpha_prod = self.alphas_cumprod[timestep]
        alpha_prod_prev = (
            self.alphas_cumprod[timestep - 1] if timestep > 0 else 1.0
        )
        beta_prod = 1 - alpha_prod
        beta_prod_prev = 1 - alpha_prod_prev

        # 2. compute predicted original sample from predicted noise also called
        # "predicted x_0" of formula (15) from https://arxiv.org/pdf/2006.11239.pdf  # noqa: E501
        if predict_epsilon:
            pred_original_sample = (
                sample - beta_prod ** (0.5) * model_output
            ) / alpha_prod ** (0.5)
        else:
            pred_original_sample = model_output

        # 3. Clip "predicted x_0"
        if self.clip_sample:
            pred_original_sample = ops.clip_by_value(
                pred_original_sample, -1, 1
            )

        # 4. Compute coefficients for pred_original_sample x_0 and current
        # sample x_t
        # See formula (7) from https://arxiv.org/pdf/2006.11239.pdf
        pred_original_sample_coeff = (
            alpha_prod_prev ** (0.5) * self.betas[timestep]
        ) / beta_prod
        current_sample_coeff = (
            self.alphas[timestep] ** (0.5) * beta_prod_prev / beta_prod
        )

        # 5. Compute predicted previous sample µ_t
        # See formula (7) from https://arxiv.org/pdf/2006.11239.pdf
        pred_prev_sample = (
            pred_original_sample_coeff * pred_original_sample
            + current_sample_coeff * sample
        )

        # 6. Add noise
        variance = 0
        if timestep > 0:
            noise = random.normal(model_output.shape, seed=self.seed_generator)
            variance = (
                self._get_variance(
                    timestep, predicted_variance=predicted_variance
                )
                ** 0.5
            ) * noise

        pred_prev_sample = pred_prev_sample + variance

        return pred_prev_sample

    def add_noise(
        self,
        original_samples,
        noise,
        timesteps,
    ):
        sqrt_alpha_prod = ops.take(self.alphas_cumprod, timesteps) ** 0.5
        sqrt_one_minus_alpha_prod = (
            1 - ops.take(self.alphas_cumprod, timesteps)
        ) ** 0.5

        for _ in range(3):
            sqrt_alpha_prod = ops.expand_dims(sqrt_alpha_prod, axis=-1)
            sqrt_one_minus_alpha_prod = ops.expand_dims(
                sqrt_one_minus_alpha_prod, axis=-1
            )
        sqrt_alpha_prod = ops.cast(
            sqrt_alpha_prod, dtype=original_samples.dtype
        )
        sqrt_one_minus_alpha_prod = ops.cast(
            sqrt_one_minus_alpha_prod, dtype=noise.dtype
        )
        noisy_samples = (
            sqrt_alpha_prod * original_samples
            + sqrt_one_minus_alpha_prod * noise
        )
        return noisy_samples

    def __len__(self):
        return self.train_timesteps



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models.feature_extractor.clip.clip_encoder import CLIPEncoder


class CLIPPatchingAndEmbedding(keras.layers.Layer):
    def __init__(
        self, width, patch_size, input_resolution, output_dim, **kwargs
    ):
        super().__init__(**kwargs)

        self.conv1 = keras.layers.Conv2D(
            filters=width,
            kernel_size=patch_size,
            strides=patch_size,
            padding="valid",
            use_bias=False,
            data_format="channels_last",
            name="patch_embed.embedding",
        )
        self.width = width
        self.input_resolution = input_resolution
        self.patch_size = patch_size
        self.num_patches = ops.power(
            (self.input_resolution // self.patch_size), 2
        )
        self.output_dim = output_dim

    def build(self, input_shape):
        super().build(input_shape)
        self.conv1.build(input_shape)
        self.class_embedding = self.add_weight(
            shape=((self.width,)),
            name="patch_embed.class_embedding",
        )

        self.positional_embedding = self.add_weight(
            shape=(
                (
                    (self.input_resolution // self.patch_size) ** 2 + 1,
                    self.width,
                )
            ),
            trainable=True,
            name="patch_embed.positional_embedding",
        )

    def compute_output_shape(self, input_shape):
        return [
            None,
            (self.input_resolution // self.patch_size) ** 2 + 1,
            self.width,
        ]

    def call(self, x):
        batch_size = ops.shape(x)[0]
        patch_embeddings = self.conv1(x)  # shape = [*, grid, grid, channel]

        patch_embeddings = ops.reshape(
            patch_embeddings, (batch_size, self.num_patches, -1)
        )
        class_embeds = ops.broadcast_to(
            self.class_embedding.value, (batch_size, 1, self.width)
        )
        embeddings = ops.concatenate(
            [class_embeds, patch_embeddings], axis=1
        )  # shape = [*, grid ** 2 + 1, width]
        positional_embedding = self.positional_embedding
        embeddings = embeddings + positional_embedding
        return embeddings

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "width": self.width,
                "patch_size": self.patch_size,
                "input_resolution": self.input_resolution,
                "output_dim": self.output_dim,
            }
        )
        return config


@keras_cv_export("keras_cv.models.feature_extractor.CLIPImageEncoder")
class CLIPImageEncoder(keras.Model):
    def __init__(
        self,
        input_resolution,
        patch_size,
        width,
        num_layers,
        heads,
        output_dim,
        **kwargs,
    ):
        super().__init__(
            **kwargs,
        )
        self.input_resolution = input_resolution
        self.width = width
        self.patch_size = patch_size
        self.output_dim = output_dim
        self.heads = heads
        self.num_layers = num_layers

        self.embeddings = CLIPPatchingAndEmbedding(
            width=self.width,
            patch_size=self.patch_size,
            input_resolution=self.input_resolution,
            output_dim=self.output_dim,
            name="clip_patch_embedding",
        )
        self.pre_norm = keras.layers.LayerNormalization(
            epsilon=1e-5, name="ln_1"
        )
        self.encoder = CLIPEncoder(
            self.width,
            self.num_layers,
            self.heads,
            name="clip_encoder",
        )
        self.post_norm = keras.layers.LayerNormalization(
            epsilon=1e-5, name="ln_2"
        )
        self.image_projector = keras.layers.Dense(
            output_dim, name="vision_projector", use_bias=False
        )

    def build(self, input_shape):
        self.embeddings.build(input_shape)
        self.pre_norm.build([None, None, self.width])
        self.encoder.build(None)
        self.post_norm.build([None, self.width])
        self.image_projector.build([None, self.width])
        self.built = True

    def compute_output_shape(self, input_shape):
        return [input_shape[0], self.output_dim]

    def call(self, image):
        x = self.embeddings(image)
        x = self.pre_norm(x)
        x = self.encoder(x)
        x = self.post_norm(x[:, 0, :])
        image_projected_embeddings = self.image_projector(x)
        return image_projected_embeddings

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_resolution": self.input_resolution,
                "patch_size": self.patch_size,
                "width": self.width,
                "layers": self.num_layers,
                "heads": self.heads,
                "output_dim": self.output_dim,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities similar to tf.python.platform.resource_loader."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import warnings

import tensorflow as tf

TF_VERSION_FOR_ABI_COMPATIBILITY = "2.13"
abi_warning_already_raised = False


def get_project_root():
    """Returns project root folder."""
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_path_to_datafile(path):
    """Get the path to the specified file in the data dependencies.

    The path is relative to keras_cv/

    Args:
      path: a string resource path relative to keras_cv/
    Returns:
      The path to the specified data file
    """
    root_dir = get_project_root()
    return os.path.join(root_dir, path.replace("/", os.sep))


class LazySO:
    def __init__(self, relative_path):
        self.relative_path = relative_path
        self._ops = None

    @property
    def ops(self):
        if self._ops is None:
            self.display_warning_if_incompatible()
            self._ops = tf.load_op_library(
                get_path_to_datafile(self.relative_path)
            )
        return self._ops

    def display_warning_if_incompatible(self):
        global abi_warning_already_raised
        if abi_warning_already_raised or abi_is_compatible():
            return

        user_version = tf.__version__
        warnings.warn(
            f"You are currently using TensorFlow {user_version} and "
            f"trying to load a KerasCV custom op.\n"
            f"KerasCV has compiled its custom ops against TensorFlow "
            f"{TF_VERSION_FOR_ABI_COMPATIBILITY}, and there are no "
            f"compatibility guarantees between the two versions.\n"
            "This means that you might get segfaults when loading the custom "
            "op, or other kind of low-level errors.\n"
            "If you do, do not file an issue on Github. "
            "This is a known limitation.",
            UserWarning,
        )
        abi_warning_already_raised = True


def abi_is_compatible():
    return tf.__version__.startswith(TF_VERSION_FOR_ABI_COMPATIBILITY)



# Copyright 2023 The KerasCV Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.point_cloud.point_cloud import _box_area
from keras_cv.src.point_cloud.point_cloud import _center_xyzWHD_to_corner_xyz
from keras_cv.src.point_cloud.point_cloud import _is_on_lefthand_side
from keras_cv.src.point_cloud.point_cloud import coordinate_transform
from keras_cv.src.point_cloud.point_cloud import group_points_by_boxes
from keras_cv.src.point_cloud.point_cloud import is_within_any_box3d
from keras_cv.src.point_cloud.point_cloud import is_within_any_box3d_v2
from keras_cv.src.point_cloud.point_cloud import is_within_any_box3d_v3
from keras_cv.src.point_cloud.point_cloud import is_within_box2d
from keras_cv.src.point_cloud.point_cloud import is_within_box3d
from keras_cv.src.point_cloud.point_cloud import spherical_coordinate_transform
from keras_cv.src.point_cloud.point_cloud import within_a_frustum
from keras_cv.src.point_cloud.point_cloud import within_box3d_index
from keras_cv.src.point_cloud.point_cloud import wrap_angle_radians



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import tensorflow as tf

from keras_cv.src.backend import ops


def to_numpy(x):
    if x is None:
        return None
    if isinstance(x, tf.RaggedTensor):
        x = x.to_tensor(-1)
    x = ops.convert_to_numpy(x)
    # Important for consistency when working with visualization utilities
    return np.ascontiguousarray(x)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.models import legacy
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetLBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetMBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetSBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetTinyBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetXLBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_backbone import (
    CSPDarkNetBackbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet121Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet169Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet201Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_backbone import (
    DenseNetBackbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (  # noqa: E501
    EfficientNetLiteB0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (  # noqa: E501
    EfficientNetLiteB1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (  # noqa: E501
    EfficientNetLiteB2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (  # noqa: E501
    EfficientNetLiteB3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (  # noqa: E501
    EfficientNetLiteB4Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_backbone import (  # noqa: E501
    EfficientNetLiteBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B4Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B5Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B6Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1B7Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (  # noqa: E501
    EfficientNetV1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2B0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2B1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2B2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2B3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2LBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2MBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (  # noqa: E501
    EfficientNetV2SBackbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB0Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB1Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB2Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB3Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB4Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTB5Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (  # noqa: E501
    MiTBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_aliases import (
    MobileNetV3LargeBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_aliases import (
    MobileNetV3SmallBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_backbone import (
    MobileNetV3Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet18Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet34Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet50Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet101Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet152Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_backbone import (
    ResNetBackbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet18V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet34V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet50V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet101V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet152V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_backbone import (
    ResNetV2Backbone,
)
from keras_cv.src.models.backbones.vgg16.vgg16_backbone import VGG16Backbone
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinBBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinSBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinTBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_backbone import (
    VideoSwinBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetBBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetHBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetLBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_backbone import (
    ViTDetBackbone,
)
from keras_cv.src.models.classification.image_classifier import ImageClassifier
from keras_cv.src.models.classification.video_classifier import VideoClassifier
from keras_cv.src.models.feature_extractor.clip import CLIP
from keras_cv.src.models.object_detection.retinanet.retinanet import RetinaNet
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_backbone import (
    YOLOV8Backbone,
)
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_detector import (
    YOLOV8Detector,
)
from keras_cv.src.models.segmentation import BASNet
from keras_cv.src.models.segmentation import DeepLabV3Plus
from keras_cv.src.models.segmentation import SAMMaskDecoder
from keras_cv.src.models.segmentation import SAMPromptEncoder
from keras_cv.src.models.segmentation import SegmentAnythingModel
from keras_cv.src.models.segmentation import TwoWayTransformer
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormer,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB0,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB1,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB2,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB3,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB4,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB5,
)
from keras_cv.src.models.stable_diffusion import StableDiffusion
from keras_cv.src.models.stable_diffusion import StableDiffusionV2



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import numpy as np

from keras_cv.src import bounding_box
from keras_cv.src import utils
from keras_cv.src.utils import assert_cv2_installed
from keras_cv.src.utils import assert_matplotlib_installed
from keras_cv.src.visualization.draw_bounding_boxes import draw_bounding_boxes
from keras_cv.src.visualization.plot_image_gallery import plot_image_gallery

try:
    from matplotlib import patches
except:
    patches = None

from keras_cv.src.api_export import keras_cv_export


@keras_cv_export("keras_cv.visualization.plot_bounding_box_gallery")
def plot_bounding_box_gallery(
    images,
    value_range,
    bounding_box_format,
    y_true=None,
    y_pred=None,
    true_color=(0, 188, 212),
    pred_color=(255, 235, 59),
    line_thickness=2,
    font_scale=1.0,
    text_thickness=None,
    class_mapping=None,
    ground_truth_mapping=None,
    prediction_mapping=None,
    legend=False,
    legend_handles=None,
    rows=3,
    cols=3,
    **kwargs
):
    """Plots a gallery of images with corresponding bounding box annotations.

    Example:
    ```python
    train_ds = tfds.load(
        "voc/2007", split="train", with_info=False, shuffle_files=True
    )

    def unpackage_tfds_inputs(inputs):
        image = inputs["image"]
        boxes = inputs["objects"]["bbox"]
        bounding_boxes = {"classes": classes, "boxes": boxes}
        return image, bounding_boxes

    train_ds = train_ds.map(unpackage_tfds_inputs)
    train_ds = train_ds.apply(tf.data.experimental.dense_to_ragged_batch(16))
    images, boxes = next(iter(train_ds.take(1)))

    keras_cv.visualization.plot_bounding_box_gallery(
        images,
        value_range=(0, 255),
        bounding_box_format="xywh",
        y_true=boxes,
        scale=3,
        rows=2,
        cols=2,
        line_thickness=4,
        font_scale=1,
        legend=True,
    )
    ```

    ![Example bounding box gallery](https://i.imgur.com/tJpb8hZ.png)

    Args:
        images: a Tensor or NumPy array containing images to show in the
            gallery.
        value_range: value range of the images. Common examples include
            `(0, 255)` and `(0, 1)`.
        bounding_box_format: the bounding_box_format the provided bounding boxes
            are in.
        y_true: (Optional) a KerasCV bounding box dictionary representing the
            ground truth bounding boxes.
        y_pred: (Optional) a KerasCV bounding box dictionary representing the
            predicted bounding boxes.
        pred_color: three element tuple representing the color to use for
            plotting predicted bounding boxes.
        true_color: three element tuple representing the color to use for
            plotting true bounding boxes.
        class_mapping: (Optional) class mapping from class IDs to strings
        ground_truth_mapping: (Optional) class mapping from class IDs to
            strings, defaults to `class_mapping`
        prediction_mapping: (Optional) class mapping from class IDs to strings,
            defaults to `class_mapping`
        line_thickness: (Optional) line_thickness for the box and text labels.
            Defaults to 2.
        text_thickness: (Optional) the line_thickness for the text, defaults to
            `1.0`.
        font_scale: (Optional) font size to draw bounding boxes in.
        legend: whether to create a legend with the specified colors for
            `y_true` and `y_pred`, defaults to False.
        kwargs: keyword arguments to propagate to
            `keras_cv.visualization.plot_image_gallery()`.
    """
    assert_matplotlib_installed("plot_bounding_box_gallery")
    assert_cv2_installed("plot_bounding_box_gallery")

    prediction_mapping = prediction_mapping or class_mapping
    ground_truth_mapping = ground_truth_mapping or class_mapping

    plotted_images = utils.to_numpy(images)

    draw_fn = functools.partial(
        draw_bounding_boxes,
        bounding_box_format="xyxy",
        line_thickness=line_thickness,
        text_thickness=text_thickness,
        font_scale=font_scale,
    )

    if y_true is not None:
        y_true = y_true.copy()
        y_true["boxes"] = utils.to_numpy(y_true["boxes"])
        y_true["classes"] = utils.to_numpy(y_true["classes"])
        y_true = bounding_box.convert_format(
            y_true, images=images, source=bounding_box_format, target="xyxy"
        )
        plotted_images = draw_fn(
            plotted_images,
            y_true,
            true_color,
            class_mapping=ground_truth_mapping,
        )

    if y_pred is not None:
        y_pred = y_pred.copy()
        y_pred["boxes"] = utils.to_numpy(y_pred["boxes"])
        y_pred["classes"] = utils.to_numpy(y_pred["classes"])
        y_pred = bounding_box.convert_format(
            y_pred, images=images, source=bounding_box_format, target="xyxy"
        )
        plotted_images = draw_fn(
            plotted_images, y_pred, pred_color, class_mapping=prediction_mapping
        )

    if legend:
        if legend_handles:
            raise ValueError(
                "Only pass `legend` OR `legend_handles` to "
                "`luketils.visualization.plot_bounding_box_gallery()`."
            )
        legend_handles = [
            patches.Patch(
                color=np.array(true_color) / 255.0,
                label="Ground Truth",
            ),
            patches.Patch(
                color=np.array(pred_color) / 255.0,
                label="Prediction",
            ),
        ]

    return plot_image_gallery(
        plotted_images,
        value_range,
        legend_handles=legend_handles,
        rows=rows,
        cols=cols,
        **kwargs
    )



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.layers.OverlappingPatchingAndEmbedding")
class OverlappingPatchingAndEmbedding(keras.layers.Layer):
    def __init__(self, project_dim=32, patch_size=7, stride=4, **kwargs):
        """
        Overlapping Patching and Embedding layer. Differs from `PatchingAndEmbedding`
        in that the patch size does not affect the sequence length. It's fully derived
        from the `stride` parameter. Additionally, no positional embedding is done
        as part of the layer - only a projection using a `Conv2D` layer.

        References:
            - [SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers](https://arxiv.org/abs/2105.15203) (CVPR 2021) # noqa: E501
            - [Official PyTorch implementation](https://github.com/NVlabs/SegFormer/blob/master/mmseg/models/backbones/mix_transformer.py) # noqa: E501
            - [Ported from the TensorFlow implementation from DeepVision](https://github.com/DavidLandup0/deepvision/blob/main/deepvision/layers/hierarchical_transformer_encoder.py) # noqa: E501

        Args:
            project_dim: integer, the dimensionality of the projection.
                Defaults to `32`.
            patch_size: integer, the size of the patches to encode.
                Defaults to `7`.
            stride: integer, the stride to use for the patching before
                projection. Defaults to `5`.

        Example:

        ```
        project_dim = 1024
        patch_size = 16

        encoded_patches = keras_cv.layers.OverlappingPatchingAndEmbedding(
        project_dim=project_dim, patch_size=patch_size)(img_batch)

        print(encoded_patches.shape) # (1, 3136, 1024)
        ```
        """
        super().__init__(**kwargs)

        self.project_dim = project_dim
        self.patch_size = patch_size
        self.stride = stride

        self.proj = keras.layers.Conv2D(
            filters=project_dim,
            kernel_size=patch_size,
            strides=stride,
            padding="same",
        )
        self.norm = keras.layers.LayerNormalization()

    def call(self, x):
        x = self.proj(x)
        # B, H, W, C
        shape = x.shape
        x = ops.reshape(x, (-1, shape[1] * shape[2], shape[3]))
        x = self.norm(x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "project_dim": self.project_dim,
                "patch_size": self.patch_size,
                "stride": self.stride,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing


@keras_cv_export("keras_cv.layers.RandomShear")
class RandomShear(VectorizedBaseImageAugmentationLayer):
    """A preprocessing layer which randomly shears images.

    This layer will apply random shearings to each image, filling empty space
    according to `fill_mode`.

    Input pixel values can be of any range and any data type.

    Input shape:
      3D (unbatched) or 4D (batched) tensor with shape:
      `(..., height, width, channels)`, in `"channels_last"` format
    Output shape:
      3D (unbatched) or 4D (batched) tensor with shape:
      `(..., height, width, channels)`, in `"channels_last"` format

    Args:
        x_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is
            sampled from the provided range. If a float is passed, the range is
            interpreted as `(0, x_factor)`. Values represent a percentage of the
            image to shear over. For example, 0.3 shears pixels up to 30% of the
            way across the image. All provided values should be positive. If
            `None` is passed, no shear occurs on the X axis. Defaults to `None`.
        y_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is
            sampled from the provided range. If a float is passed, the range is
            interpreted as `(0, y_factor)`. Values represent a percentage of the
            image to shear over. For example, 0.3 shears pixels up to 30% of the
            way across the image. All provided values should be positive. If
            `None` is passed, no shear occurs on the Y axis. Defaults to `None`.
        interpolation: interpolation method used in the
            `ImageProjectiveTransformV3` op. Supported values are `"nearest"`
            and `"bilinear"`, defaults to `"bilinear"`.
        fill_mode: fill_mode in the `ImageProjectiveTransformV3` op. Supported
            values are `"reflect"`, `"wrap"`, `"constant"`, and `"nearest"`.
            Defaults to `"reflect"`.
        fill_value: fill_value in the `ImageProjectiveTransformV3` op. A
            `Tensor` of type `float32`. The value to be filled when fill_mode is
            constant". Defaults to `0.0`.
        bounding_box_format: The format of bounding boxes of input dataset.
            Refer to
            https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box/converters.py
            for more details on supported bounding box formats.
        seed: Integer. Used to create a random seed.
    """

    def __init__(
        self,
        x_factor=None,
        y_factor=None,
        interpolation="bilinear",
        fill_mode="reflect",
        fill_value=0.0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if x_factor is not None:
            self.x_factor = preprocessing.parse_factor(
                x_factor, max_value=None, param_name="x_factor", seed=seed
            )
        else:
            self.x_factor = x_factor
        if y_factor is not None:
            self.y_factor = preprocessing.parse_factor(
                y_factor, max_value=None, param_name="y_factor", seed=seed
            )
        else:
            self.y_factor = y_factor
        if x_factor is None and y_factor is None:
            warnings.warn(
                "RandomShear received both `x_factor=None` and `y_factor=None`."
                " As a result, the layer will perform no augmentation."
            )
        self.interpolation = interpolation
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.seed = seed
        self.bounding_box_format = bounding_box_format

    def get_random_transformation_batch(self, batch_size, **kwargs):
        transformations = {"shear_x": None, "shear_y": None}
        if self.x_factor is not None:
            invert = preprocessing.batch_random_inversion(
                self._random_generator, batch_size
            )
            transformations["shear_x"] = (
                self.x_factor(shape=(batch_size, 1)) * invert
            )

        if self.y_factor is not None:
            invert = preprocessing.batch_random_inversion(
                self._random_generator, batch_size
            )
            transformations["shear_y"] = (
                self.y_factor(shape=(batch_size, 1)) * invert
            )

        return transformations

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        new_transformation = {"shear_x": None, "shear_y": None}
        shear_x = transformation["shear_x"]
        if shear_x is not None:
            new_transformation["shear_x"] = tf.expand_dims(shear_x, axis=0)

        shear_y = transformation["shear_y"]
        if shear_y is not None:
            new_transformation["shear_y"] = tf.expand_dims(shear_y, axis=0)

        output = self.augment_images(images, new_transformation)
        return tf.squeeze(output, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        x, y = transformations["shear_x"], transformations["shear_y"]

        if x is not None:
            transforms_x = self._build_shear_x_transform_matrix(x)
            images = preprocessing.transform(
                images=images,
                transforms=transforms_x,
                interpolation=self.interpolation,
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        if y is not None:
            transforms_y = self._build_shear_y_transform_matrix(y)
            images = preprocessing.transform(
                images=images,
                transforms=transforms_y,
                interpolation=self.interpolation,
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        return images

    @staticmethod
    def _build_shear_x_transform_matrix(shear_x):
        """Build transform matrix for horizontal shear.

        The transform matrix looks like:
        (1, x, 0)
        (0, 1, 0)
        (0, 0, 1)
        where the last entry is implicit.

        We flatten the matrix to `[1.0, x, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]` for
        use with ImageProjectiveTransformV3.
        """
        batch_size = tf.shape(shear_x)[0]
        return tf.concat(
            values=[
                tf.ones((batch_size, 1), tf.float32),
                shear_x,
                tf.zeros((batch_size, 2), tf.float32),
                tf.ones((batch_size, 1), tf.float32),
                tf.zeros((batch_size, 3), tf.float32),
            ],
            axis=1,
        )

    @staticmethod
    def _build_shear_y_transform_matrix(shear_y):
        """Build transform matrix for vertical shear.

        The transform matrix looks like:
        (1, 0, 0)
        (y, 1, 0)
        (0, 0, 1)
        where the last entry is implicit.

        We flatten the matrix to `[1.0, 0.0, 0.0, y, 1.0, 0.0, 0.0, 0.0]` for
        use ImageProjectiveTransformV3.
        """
        batch_size = tf.shape(shear_y)[0]
        return tf.concat(
            values=[
                tf.ones((batch_size, 1), tf.float32),
                tf.zeros((batch_size, 2), tf.float32),
                shear_y,
                tf.ones((batch_size, 1), tf.float32),
                tf.zeros((batch_size, 3), tf.float32),
            ],
            axis=1,
        )

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        x, y = transformations["shear_x"], transformations["shear_y"]

        if x is not None:
            transforms_x = self._build_shear_x_transform_matrix(x)
            segmentation_masks = preprocessing.transform(
                images=segmentation_masks,
                transforms=transforms_x,
                interpolation="nearest",
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        if y is not None:
            transforms_y = self._build_shear_y_transform_matrix(y)
            segmentation_masks = preprocessing.transform(
                images=segmentation_masks,
                transforms=transforms_y,
                interpolation="nearest",
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        return segmentation_masks

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, images=None, **kwargs
    ):
        """Augments bounding boxes after a shear operations.

        The algorithm to update (x,y) point coordinates after shearing, tells us
        to matrix multiply them with inverted transform matrix. This is:
        ```
        # for shear x              # for shear_y
        (1.0, -shear_x) (x)        (1.0,      0.0) (x)
        (0.0, 1.0     ) (y)        (-shear_y, 1.0) (y)
        ```
        We can simplify this equation: any new coordinate can be calculated by
        `x = x - (shear_x * y)` and `(y = y - (shear_y * x)`

        Notice that each coordinate has to be calculated twice, e.g. `x1` will
        be affected differently by y1 (top) and y2 (bottom). Therefore, we
        calculate both `x1_top` and `x1_bottom` and choose the final x1
        depending on the sign of the used shear value.
        """
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomShear()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomShear(bounding_box_format='xyxy')`"
            )

        # Edge case: boxes is a tf.RaggedTensor
        if isinstance(bounding_boxes["boxes"], tf.RaggedTensor):
            bounding_boxes = bounding_box.to_dense(
                bounding_boxes, default_value=0
            )

        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="rel_xyxy",
            images=images,
            dtype=self.compute_dtype,
        )

        shear_x_amount = transformations["shear_x"]
        shear_y_amount = transformations["shear_y"]
        x1, y1, x2, y2 = tf.split(bounding_boxes["boxes"], 4, axis=-1)

        # Squeeze redundant extra dimension as it messes multiplication
        # [num_batches, num_boxes, 1] -> [num_batches, num_boxes]
        x1 = tf.squeeze(x1, axis=-1)
        y1 = tf.squeeze(y1, axis=-1)
        x2 = tf.squeeze(x2, axis=-1)
        y2 = tf.squeeze(y2, axis=-1)

        # Apply horizontal shear
        if shear_x_amount is not None:
            x1_top = x1 - (shear_x_amount * y1)
            x1_bottom = x1 - (shear_x_amount * y2)
            x1 = tf.where(shear_x_amount < 0, x1_top, x1_bottom)

            x2_top = x2 - (shear_x_amount * y1)
            x2_bottom = x2 - (shear_x_amount * y2)
            x2 = tf.where(shear_x_amount < 0, x2_bottom, x2_top)

        # Apply vertical shear
        if shear_y_amount is not None:
            y1_left = y1 - (shear_y_amount * x1)
            y1_right = y1 - (shear_y_amount * x2)
            y1 = tf.where(shear_y_amount > 0, y1_right, y1_left)

            y2_left = y2 - (shear_y_amount * x1)
            y2_right = y2 - (shear_y_amount * x2)
            y2 = tf.where(shear_y_amount > 0, y2_left, y2_right)

        # Join the results:
        boxes = tf.concat(
            [
                # Add dummy last axis for concat:
                # (num_batches, num_boxes) -> (num_batches, num_boxes, 1)
                x1[..., tf.newaxis],
                y1[..., tf.newaxis],
                x2[..., tf.newaxis],
                y2[..., tf.newaxis],
            ],
            axis=-1,
        )

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes, images=images, bounding_box_format="rel_xyxy"
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="rel_xyxy",
            target=self.bounding_box_format,
            images=images,
            dtype=self.compute_dtype,
        )
        return bounding_boxes

    @staticmethod
    def _format_transform(transform):
        transform = tf.convert_to_tensor(transform, dtype=tf.float32)
        return transform[tf.newaxis]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "x_factor": self.x_factor,
                "y_factor": self.y_factor,
                "interpolation": self.interpolation,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras


@keras_cv_export("keras_cv.layers.Augmenter")
class Augmenter(keras.layers.Layer):
    """Light-weight class to apply augmentations to data.

    Args:
        layers: A list of `keras.layers.Layers` to apply to the example

    Example:
    from keras_cv.src import layers
    images = np.ones((16, 256, 256, 3))
    augmenter = layers.Augmenter(
        [
            layers.RandomFlip(),
            layers.RandAugment(value_range=(0, 255)),
            layers.CutMix(),
        ]
    )
    augmented_images = augmenter(images)
    """

    def __init__(self, layers):
        self.layers = layers

    def __call__(self, inputs):
        for layer in self.layers:
            inputs = layer(inputs)
        return inputs



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.api.models import classification
from keras_cv.api.models import feature_extractor
from keras_cv.api.models import object_detection
from keras_cv.api.models import retinanet
from keras_cv.api.models import segmentation
from keras_cv.api.models import stable_diffusion
from keras_cv.api.models import yolov8
from keras_cv.src.models.backbones.backbone import Backbone
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetLBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetMBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetSBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetTinyBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_aliases import (
    CSPDarkNetXLBackbone,
)
from keras_cv.src.models.backbones.csp_darknet.csp_darknet_backbone import (
    CSPDarkNetBackbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet121Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet169Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_aliases import (
    DenseNet201Backbone,
)
from keras_cv.src.models.backbones.densenet.densenet_backbone import (
    DenseNetBackbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (
    EfficientNetLiteB0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (
    EfficientNetLiteB1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (
    EfficientNetLiteB2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (
    EfficientNetLiteB3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_aliases import (
    EfficientNetLiteB4Backbone,
)
from keras_cv.src.models.backbones.efficientnet_lite.efficientnet_lite_backbone import (
    EfficientNetLiteBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B4Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B5Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B6Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_aliases import (
    EfficientNetV1B7Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v1.efficientnet_v1_backbone import (
    EfficientNetV1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2B0Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2B1Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2B2Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2B3Backbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2LBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2MBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_aliases import (
    EfficientNetV2SBackbone,
)
from keras_cv.src.models.backbones.efficientnet_v2.efficientnet_v2_backbone import (
    EfficientNetV2Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB0Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB1Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB2Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB3Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB4Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_aliases import (
    MiTB5Backbone,
)
from keras_cv.src.models.backbones.mix_transformer.mix_transformer_backbone import (
    MiTBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_aliases import (
    MobileNetV3LargeBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_aliases import (
    MobileNetV3SmallBackbone,
)
from keras_cv.src.models.backbones.mobilenet_v3.mobilenet_v3_backbone import (
    MobileNetV3Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet18Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet34Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet50Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet101Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_aliases import (
    ResNet152Backbone,
)
from keras_cv.src.models.backbones.resnet_v1.resnet_v1_backbone import (
    ResNetBackbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet18V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet34V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet50V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet101V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_aliases import (
    ResNet152V2Backbone,
)
from keras_cv.src.models.backbones.resnet_v2.resnet_v2_backbone import (
    ResNetV2Backbone,
)
from keras_cv.src.models.backbones.vgg16.vgg16_backbone import VGG16Backbone
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinBBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinSBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_aliases import (
    VideoSwinTBackbone,
)
from keras_cv.src.models.backbones.video_swin.video_swin_backbone import (
    VideoSwinBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetBBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetHBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_aliases import (
    ViTDetLBackbone,
)
from keras_cv.src.models.backbones.vit_det.vit_det_backbone import (
    ViTDetBackbone,
)
from keras_cv.src.models.classification.image_classifier import ImageClassifier
from keras_cv.src.models.classification.video_classifier import VideoClassifier
from keras_cv.src.models.feature_extractor.clip.clip_model import CLIP
from keras_cv.src.models.object_detection.retinanet.retinanet import RetinaNet
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_backbone import (
    YOLOV8Backbone,
)
from keras_cv.src.models.object_detection.yolo_v8.yolo_v8_detector import (
    YOLOV8Detector,
)
from keras_cv.src.models.object_detection_3d.center_pillar import (
    MultiHeadCenterPillar,
)
from keras_cv.src.models.object_detection_3d.center_pillar_backbone import (
    CenterPillarBackbone,
)
from keras_cv.src.models.segmentation.basnet.basnet import BASNet
from keras_cv.src.models.segmentation.deeplab_v3_plus.deeplab_v3_plus import (
    DeepLabV3Plus,
)
from keras_cv.src.models.segmentation.segformer.segformer import SegFormer
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB0,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB1,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB2,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB3,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB4,
)
from keras_cv.src.models.segmentation.segformer.segformer_aliases import (
    SegFormerB5,
)
from keras_cv.src.models.segmentation.segment_anything.sam import (
    SegmentAnythingModel,
)
from keras_cv.src.models.segmentation.segment_anything.sam_mask_decoder import (
    SAMMaskDecoder,
)
from keras_cv.src.models.segmentation.segment_anything.sam_prompt_encoder import (
    SAMPromptEncoder,
)
from keras_cv.src.models.segmentation.segment_anything.sam_transformer import (
    TwoWayTransformer,
)
from keras_cv.src.models.stable_diffusion.stable_diffusion import (
    StableDiffusion,
)
from keras_cv.src.models.stable_diffusion.stable_diffusion import (
    StableDiffusionV2,
)
from keras_cv.src.models.task import Task



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MobileNetV3 model preset configurations."""

backbone_presets_no_weights = {
    "mobilenet_v3_small": {
        "metadata": {
            "description": (
                "MobileNetV3 model with 14 layers where the batch "
                "normalization and hard-swish activation are applied after the "
                "convolution layers."
            ),
            "params": 933502,
            "official_name": "MobileNetV3",
            "path": "mobilenetv3",
        },
        "kaggle_handle": "kaggle://keras/mobilenetv3/keras/mobilenet_v3_small/2",  # noqa: E501
    },
    "mobilenet_v3_large": {
        "metadata": {
            "description": (
                "MobileNetV3 model with 28 layers where the batch "
                "normalization and hard-swish activation are applied after the "
                "convolution layers."
            ),
            "params": 2994518,
            "official_name": "MobileNetV3",
            "path": "mobilenetv3",
        },
        "kaggle_handle": "kaggle://keras/mobilenetv3/keras/mobilenet_v3_large/2",  # noqa: E501
    },
}

backbone_presets_with_weights = {
    "mobilenet_v3_large_imagenet": {
        "metadata": {
            "description": (
                "MobileNetV3 model with 28 layers where the batch "
                "normalization and hard-swish activation are applied after the "
                "convolution layers. "
                "Pre-trained on the ImageNet 2012 classification task."
            ),
            "params": 2994518,
            "official_name": "MobileNetV3",
            "path": "mobilenetv3",
        },
        "kaggle_handle": "kaggle://keras/mobilenetv3/keras/mobilenet_v3_large_imagenet/2",  # noqa: E501
    },
    "mobilenet_v3_small_imagenet": {
        "metadata": {
            "description": (
                "MobileNetV3 model with 14 layers where the batch "
                "normalization and hard-swish activation are applied after the "
                "convolution layers. "
                "Pre-trained on the ImageNet 2012 classification task."
            ),
            "params": 933502,
            "official_name": "MobileNetV3",
            "path": "mobilenetv3",
        },
        "kaggle_handle": "kaggle://keras/mobilenetv3/keras/mobilenet_v3_small_imagenet/2",  # noqa: E501
    },
}

backbone_presets = {
    **backbone_presets_no_weights,
    **backbone_presets_with_weights,
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DenseNet model preset configurations."""

backbone_presets_no_weights = {
    "densenet121": {
        "metadata": {
            "description": "DenseNet model with 121 layers.",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet121/2",
    },
    "densenet169": {
        "metadata": {
            "description": "DenseNet model with 169 layers.",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet169/2",
    },
    "densenet201": {
        "metadata": {
            "description": "DenseNet model with 201 layers.",
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet201/2",
    },
}

backbone_presets_with_weights = {
    "densenet121_imagenet": {
        "metadata": {
            "description": (
                "DenseNet model with 121 layers. Trained on Imagenet 2012 "
                "classification task."
            ),
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet121_imagenet/2",
    },
    "densenet169_imagenet": {
        "metadata": {
            "description": (
                "DenseNet model with 169 layers. Trained on Imagenet 2012 "
                "classification task."
            ),
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet169_imagenet/2",
    },
    "densenet201_imagenet": {
        "metadata": {
            "description": (
                "DenseNet model with 201 layers. Trained on Imagenet 2012 "
                "classification task."
            ),
        },
        "kaggle_handle": "kaggle://keras/densenet/keras/densenet201_imagenet/2",
    },
}

backbone_presets = {
    **backbone_presets_no_weights,
    **backbone_presets_with_weights,
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl.testing import parameterized

from keras_cv.src.backend import ops


class TestCase(tf.test.TestCase, parameterized.TestCase):
    """Base test case class for KerasCV. (Copied from KerasNLP)."""

    def assertAllClose(self, x1, x2, atol=1e-6, rtol=1e-6, msg=None):
        x1 = tf.nest.map_structure(convert_to_numpy, x1)
        x2 = tf.nest.map_structure(convert_to_numpy, x2)
        super().assertAllClose(x1, x2, atol=atol, rtol=rtol, msg=msg)

    def assertAllEqual(self, x1, x2, msg=None):
        x1 = tf.nest.map_structure(convert_to_numpy, x1)
        x2 = tf.nest.map_structure(convert_to_numpy, x2)
        super().assertAllEqual(x1, x2, msg=msg)

    def assertAllGreaterEqual(self, x1, x2):
        x1 = tf.nest.map_structure(convert_to_numpy, x1)
        x2 = tf.nest.map_structure(convert_to_numpy, x2)
        super().assertAllGreaterEqual(x1, x2)

    def assertAllLessEqual(self, x1, x2):
        x1 = tf.nest.map_structure(convert_to_numpy, x1)
        x2 = tf.nest.map_structure(convert_to_numpy, x2)
        super().assertAllLessEqual(x1, x2)


def convert_to_numpy(x):
    if ops.is_tensor(x) and not isinstance(x, tf.RaggedTensor):
        return ops.convert_to_numpy(x)
    return x



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    # When using torch and tensorflow, torch needs to be imported first,
    # otherwise it will segfault upon import.
    import torch

    del torch
except ImportError:
    pass

# isort:off
from keras_cv.src import version_check

version_check.check_tf_version()
# isort:on

from keras_cv.src import bounding_box  # noqa: E402
from keras_cv.src import callbacks  # noqa: E402
from keras_cv.src import datasets  # noqa: E402
from keras_cv.src import layers  # noqa: E402
from keras_cv.src import losses  # noqa: E402
from keras_cv.src import metrics  # noqa: E402
from keras_cv.src import models  # noqa: E402
from keras_cv.src import training  # noqa: E402
from keras_cv.src import utils  # noqa: E402
from keras_cv.src import visualization  # noqa: E402
from keras_cv.src.core import ConstantFactorSampler  # noqa: E402
from keras_cv.src.core import FactorSampler  # noqa: E402
from keras_cv.src.core import NormalFactorSampler  # noqa: E402
from keras_cv.src.core import UniformFactorSampler  # noqa: E402
from keras_cv.src.version_utils import __version__  # noqa: E402
from keras_cv.src.version_utils import version  # noqa: E402



# Copyright 2022 The KerasCV Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.models.stable_diffusion.padded_conv2d import PaddedConv2D


class AttentionBlock(keras.layers.Layer):
    def __init__(self, output_dim, **kwargs):
        super().__init__(**kwargs)
        self.output_dim = output_dim
        self.norm = keras.layers.GroupNormalization(epsilon=1e-5)
        self.q = PaddedConv2D(output_dim, 1)
        self.k = PaddedConv2D(output_dim, 1)
        self.v = PaddedConv2D(output_dim, 1)
        self.proj_out = PaddedConv2D(output_dim, 1)

    def call(self, inputs):
        x = self.norm(inputs)
        q, k, v = self.q(x), self.k(x), self.v(x)

        # Compute attention
        shape = ops.shape(q)
        h, w, c = shape[1], shape[2], shape[3]
        q = ops.reshape(q, (-1, h * w, c))  # b, hw, c
        k = ops.transpose(k, (0, 3, 1, 2))
        k = ops.reshape(k, (-1, c, h * w))  # b, c, hw
        y = q @ k
        y = y * 1 / ops.sqrt(ops.cast(c, self.compute_dtype))
        y = keras.activations.softmax(y)

        # Attend to values
        v = ops.transpose(v, (0, 3, 1, 2))
        v = ops.reshape(v, (-1, c, h * w))
        y = ops.transpose(y, (0, 2, 1))
        x = v @ y
        x = ops.transpose(x, (0, 2, 1))
        x = ops.reshape(x, (-1, h, w, c))
        return self.proj_out(x) + inputs
